/**
 *  $Id: HLClient.java,v 1.17 2001/10/05 14:19:07 groomed Exp $ Copyright (C)
 *  1998-2001 groomed <groomed@users.sourceforge.net> This program is free
 *  software; you can redistribute it and/or modify it under the terms of the
 *  GNU General Public License as published by the Free Software Foundation;
 *  either version 2 of the License, or (at your option) any later version. This
 *  program is distributed in the hope that it will be useful, but WITHOUT ANY
 *  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 *  details. You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package pharerouge.hotline;

//{{{ imports
import java.io.BufferedOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.PipedInputStream;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;
import java.util.Vector;

import pharerouge.utils.DebuggerOutput;
import pharerouge.utils.ToArrayConverters;
//}}}

//{{{ HLClient(String) javadoc 
/**
 *  This class provides a high-level API for communicating with Hotline servers.
 *  <P>
 *
 *  Let's start out with a simple example:<P>
 *
 *  <BLOCKQUOTE> <PRE>
 * <FONT color="#ff0000">// pharerouge.hotline.* example program.
 * //
 * // This program demonstrates how to connect to a Hotline
 * // server and show some bad manners.</FONT> import pharerouge.hotline.*; import
 *  java.io.*; public class HLDemo { public static void main(String args[]) { HL
 *  h = new HL("hlserver.com"); } } <FONT color="#ff0000"> // We extend {@link
 *  pharerouge.hotline.HLClientAdapter} so we can register // ourselves as a
 *  {@link pharerouge.hotline.HLClientListener} to receive // events from the
 *  server. </FONT> class HL extends {@link pharerouge.hotline.HLClientAdapter} {
 *  HLClient hlc; HL(String server) { try { hlc = new HLClient(server, <FONT
 *  color="#ff0000">// Server address.</FONT> {@link
 *  pharerouge.hotline.HLProtocol#HTLS_TCPPORT}, <FONT color="#ff0000">// Port
 *  number.</FONT> "guest", <FONT color="#ff0000">// Login.</FONT> "", <FONT
 *  color="#ff0000">// Password.</FONT> "HLDemo", <FONT color="#ff0000">//
 *  Nickname.</FONT> (short) 0); <FONT color="#ff0000">// Icon number.</FONT>
 *  <FONT color="#ff0000"> // This is where we register ourselves as the handler
 *  for // server events.</FONT> hlc.addHLClientListener(this); <FONT
 *  color="#ff0000">// Open the connection to the server.</FONT> hlc.connect();
 *  <FONT color="#ff0000">// Send the user information to the server.</FONT>
 *  hlc.login(); <FONT color="#ff0000">// Show off.</FONT> demo1(); demo2();
 *  demo3(); <FONT color="#ff0000">// Stay alive for 20 seconds.</FONT>
 *  Thread.currentThread().sleep(20000); } catch(InterruptedException e) { <FONT
 *  color="#ff0000">// May be thrown by the sleep() call above.</FONT> }
 *  catch(FailedLoginException e) { <FONT color="#ff0000">// May be thrown from
 *  {@link #login}.</FONT> System.out.println("Could not login to the server: "
 *  + e.getMessage()); } catch(IOException e) { <FONT color="#ff0000">// May be
 *  thrown from {@link #connect} or {@link #login}.</FONT> System.out.println("A
 *  network error occurred: " + e.getMessage()); } catch(HLException e) { <FONT
 *  color="#ff0000">// May be thrown in some of the HLClient methods.</FONT>
 *  System.out.println("A Hotline server error occurred: " + e.getMessage()); }
 *  finally { try { hlc.close(); } catch(IOException e) {} } } <FONT
 *  color="#ff0000">// Retrieve the contents of the root folder.</FONT> void
 *  demo1() throws IOException, {@link pharerouge.hotline.HLException} { {@link
 *  HLProtocol.FileListComponent}[] files = hlc.getFileList(":"); for(int i=0;
 *  i&lt;files.length; i++) System.out.println(new String(files[i].fname)); }
 *  <FONT color="#ff0000">// Send everybody a message.</FONT> void demo2()
 *  throws IOException, {@link pharerouge.hotline.HLException} { {@link
 *  HLProtocol.UserListComponent}[] users = hlc.getUserList(); for(int i=0;
 *  i&lt;users.length; i++) hlc.sendMessage(users[i].sock, "Hi there!"); } <FONT
 *  color="#ff0000">// Write silly text to chat.</FONT> void demo3() throws
 *  IOException { hlc.sendChat("I am silly"); } <FONT color="#ff0000">//
 *  Inherited from {@link pharerouge.hotline.HLClientAdapter}. // This method is
 *  invoked when chatting is received.</FONT> public void handleChat(String msg)
 *  { System.out.println("chat received: "+msg); } <FONT color="#ff0000">//
 *  Inherited from {@link pharerouge.hotline.HLClientAdapter}. // This method is
 *  invoked when a message is received.</FONT> public void handleMessage(int s,
 *  String n, String msg) { System.out.println("message from "+n+": "+msg); try
 *  { <FONT color="#ff0000">// Return the message that was received to the
 *  sender.</FONT> hlc.sendMessage(s, msg); } catch(IOException e) {} } }</PRE>
 *  </BLOCKQUOTE> <B>Tasks and transaction IDs</B> <P>
 *
 *  A task is a client-initiated transaction. Examples of tasks are logging in,
 *  sending a private message and getting a filelist. Hotline identifies
 *  different tasks using a transaction ID. <p>
 *
 *  This transaction ID can be used for example to wait for completion of a
 *  task, using the {@link #waitFor} method. The {@link #waitFor} method blocks
 *  until a reply has been received for the task identified by the transaction
 *  ID. The contents of the task reply is then returned as a generic Object,
 *  which must be typecast to the desired type depending on the task performed.
 *  For this reason, a number of convenience functions (with the prefix <i>get
 *  </i>) are provided, which perform the typecasting for you automatically.<p>
 *
 *  Note that Hotline servers do not send a reply for chat tasks (that is, those
 *  created using the {@link #sendChat} and {@link #sendChatAction} methods)
 *  unless an error occurs.<p>
 *
 *  <B>Retrieving information: get vs. request</B> <P>
 *
 *  Some methods exist in both a <TT>request</TT> and a <TT>get</TT> variant.
 *  The <TT>get</TT> version fetches the requested information from the server
 *  and returns it. The <TT>request</TT> version, by contrast, only requests the
 *  information something and returns immediately. This leaves handling up to
 *  any registered HLClientListeners. <P>
 *
 *  The <TT>get</TT> version will block until the data has been received and
 *  return it. This is frequently convenient, because it bypasses the need to
 *  register an HLClientListener.<P>
 *
 *  Note that the <TT>get</TT> variants can return <TT>null</TT> when an error
 *  occurs. Also, they throw an exception when the connection to the server is
 *  closed (either by using the {@link #close} method, or because the connection
 *  with the server was terminated).<P>
 *
 *  <B>Files and pathname conventions</B> <P>
 *
 *  Hotline is originally a Macintosh application. This has some implications
 *  for the way directories and files are handled.<P>
 *
 *  Instead of using the familiar slash (/) to separate directories, the Mac
 *  uses a colon (:). Thus, a Hotline path to the file <TT>Fubar</TT> might be
 *  <TT>"Foo:Bar:Fubar"</TT> . For maximum flexibility you can use the {@link
 *  HLProtocol#DIR_SEPARATOR} constant instead of :.<P>
 *
 *  Also, the Mac concept of a file is different from Windows and Unix
 *  platforms. Most importantly, Mac files differ in that they can contain two
 *  separate data streams, called the data fork and the resource fork. This does
 *  not easily translate to Java's file flat-file model. In addition, the
 *  Macintosh filesystem tracks some meta-data about files that simply does not
 *  exist on other platforms, such as type and creator codes. <p>
 *
 *  Therefore, the file transfer API uses {@link pharerouge.macfiles.MacFile}
 *  objects to represent the local file. The application can choose between a
 *  number of strategies for dealing with these discrepancies by instantiating
 *  the appropriate {@link pharerouge.macfiles.MacFile} subclass.<p>
 *
 *  <B>Newlines</B> <P>
 *
 *  Again, Hotline is originally a Macintosh application. The Macintosh uses \r
 *  (ASCII 0x0d) to indicate a line ending, while Java has adopted the Unix
 *  convention of \n (ASCII 0x0a) for line endings. This means that text
 *  received from the server (e.g. messages, chat, news, file comments) may need
 *  to have have \r replaced by \n. You can use the {@link
 *  pharerouge.utils.TextUtils#findAndReplace TextUtils.findAndReplace} function
 *  to do this, e.g. <BR>
 *  <BLOCKQUOTE> <PRE>
 * String converted = TextUtils.findAndReplace(someString, "\r", "\n");</PRE>
 *  </BLOCKQUOTE> Conversely, when sending text to a Hotline server, you may
 *  need to replace \n by \r.<P>
 *
 *  <b>Threading and event handling</b> <p>
 *
 *  A HLClient uses a seperate thread for reading packets from the server and a
 *  seperate thread for writing packets to the server. In addition, every packet
 *  read from the server is dispatched in a separate dispatcher thread.<p>
 *
 *  To terminate all these running threads, it is important to always call the
 *  {@link #close} method when you are done.<p>
 *
 *  Most of the HLClient object methods are thread-safe, with the notable
 *  exception of {@link #connect} and {@link #login}. Generally thread-safe, but
 *  with some restrictions, is {@link #setBlocked}.<p>
 *
 *  <B>File transfer</B> <P>
 *
 *  Transferring files is done using the {@link #requestFileUpload} and {@link
 *  #requestFileDownload} methods. Both methods share three common parameters:
 *
 *  <UL>
 *    <LI> a {@link pharerouge.macfiles.MacFile} object specifying the local file
 *    </LI>
 *    <LI> a String specifying the remote file/path</LI>
 *    <LI> a {@link pharerouge.utils.Meter} object that monitors the transfer
 *    progress</LI>
 *  </UL>
 *  <U>The <TT>MacFile</TT> object</U> <P>
 *
 *  The MacFile object describes a Macintosh file, which can contain information
 *  that "conventional" (e.g. Unix) files do not have.<P>
 *
 *  Creating a MacFile from an ordinary File is accomplished simply by passing
 *  the File as an argument to the MacFile constructor:<P>
 *
 *  <BLOCKQUOTE><PRE>
 * File f = new File("thefile");
 * MacBinaryMacFile mf = new MacBinaryMacFile(f); // Can also be SplitMacFile or NativeMacFile.
 * </PRE></BLOCKQUOTE> Note that a MacFile may actually consist of multiple
 *  files on disk or have it's name changed slightly (e.g. a {@link
 *  pharerouge.macfiles.MacBinaryMacFile} creates a file with ".bin" appended to
 *  the filename). See the MacFile documentation for more info. <P>
 *
 *  <U>The <TT>Meter</TT> object</U> <P>
 *
 *  The <TT>Meter</TT> object monitors the transfer progress. Interrupting a
 *  transfer is also done through the <TT>Meter</TT> object. For example:<P>
 *
 *  <BLOCKQUOTE> <PRE>
 * <FONT color="#ff0000">
 * // A simple example of a Meter object that monitors a transfer.
 * // The stopTransfer method demonstrates how to stop a transfer.</FONT> public
 *  class TransferMeter implements {@link pharerouge.utils.Meter} { {@link
 *  pharerouge.utils.MeterSource} transferrer; void stopTransfer() { <FONT
 *  color="#ff0000">// Interrupts the transfer.</FONT> transferrer.interrupt();
 *  } <FONT color="#ff0000">// Following methods implement Meter.</FONT> public
 *  void startMeter(String s, int size) { System.out.println("transfer started,
 *  file = "+s+", size = "+size); } public void progressMeter(int done) {
 *  System.out.println("transfer progress = "+done); } public void stopMeter() {
 *  System.out.println("transfer completed successfully"); } public void
 *  stopMeterWithError(Throwable t) { System.out.println("transfer completed
 *  abnormally: "+t.getMessage()); } public void setMeterSource({@link
 *  pharerouge.utils.MeterSource} ms) { transferrer = ms; } public MeterSource
 *  getMeterSource() { return transferrer; } } </PRE> </BLOCKQUOTE> <P>
 *
 *  Note that you cannot depend on the {@link #close} method to stop transfers
 *  that are in progress. A file transfer <i>has to be</i> stopped by calling
 *  the {@link pharerouge.utils.MeterSource#interrupt} method from the {@link
 *  pharerouge.utils.Meter} that monitors that transfer (see the stopTransfer
 *  method in the example above). <P>
 *
 *
 */
//}}}
public class HLClient {

    /* Communications. */
    InetAddress host;
    int port;
    Socket socket;
    DataOutputStream output;
    DataInputStream input;
    ReaderThread reader;
    HLClientDispatcher hlr;

    // 2002.10.25 MackyBee EXPERIMENTALLY modified.
    Map tasks;
    Map transfers;

    //Hashtable tasks, transfers;

    // 2002.09.09 MackyBee added a Javadoc comment.
    /* Vector containing the event listeners for this client. */
    /** Contains HLClientListener object reference variables. */
    Vector hle;

    String nick = default_nick;
    String login = default_login;
    String password = default_password;

    /* Modified on 2002.03.18 by MackyBee
	   The icon should be 'short'. */
    short icon = default_icon;

    /* Internal state flags and communication variables. */
    boolean connected = false, blocked = false;
    /* 2002.04.13 MackyBee modified */
    int transaction = 1;
    int serverVersion = 0;
    //int transaction = 2, serverVersion = 0;
    
    // 2003.01.30 MackyBee added. 
    String serverName;
    
    Integer closeLock = new Integer(0);

    /* Create these objects once instead of creating one for every
       transaction. */
    HLProtocol hlp;

    /* Setup defaults for login data. */
    private static String default_nick = "PhareRouge library";
    private static String default_login = "";
    private static String default_password = "";
    private static short default_icon = 2015;
    //int icon = default_icon;

    // 2002.03.25 MackyBee added experimentally
    private boolean loginSucceeded = false;
    
    /** Character encoding */
    // 2002.12.03 MackyBee added. 
    String encoding = "ISO-8859-1";

    /**
     *  Prepare a connection to the specified server.
     *
     * @param  server           the server name to log into.
     * @exception  IOException  Description of the Exception
     */
    public HLClient(String server) throws IOException {
        this(InetAddress.getByName(server), HLProtocol.HTLS_TCPPORT);
    }

    /**
     *  Prepare a connection to the specified host address.
     *
     * @param  host             the IP address of the host to log into.
     * @exception  IOException  Description of the Exception
     */
    public HLClient(InetAddress host) throws IOException {
        this(host, HLProtocol.HTLS_TCPPORT);
    }

    /**
     *  Prepare a connection to the specified server at the specified port.
     *
     * @param  server           the server name
     * @param  port             the port number
     * @exception  IOException  Description of the Exception
     */
    public HLClient(String server, int port) throws IOException {

        this(InetAddress.getByName(server),
            port,
            default_login,
            default_password,
            default_nick,
            default_icon);
    }

    /**
     *  Prepare a connection to the specified host at the specified port.
     *
     * @param  host             the IP address of the host to log into.
     * @param  port             the port number
     * @exception  IOException  Description of the Exception
     */
    public HLClient(InetAddress host, int port) throws IOException {

        this(host,
            port,
            default_login,
            default_password,
            default_nick,
            default_icon);
    }

    /**
     *  Prepare a connection to the specified server at the specified port with
     *  specified user credentials. Usually "guest" and "" (empty string) will
     *  suffice as credentials.
     *
     * @param  server                    the server name
     * @param  port                      the port number
     * @param  login                     the hotline user login
     * @param  password                  the hotline user password
     * @param  nick                      the hotline user nickname
     * @param  icon                      the hotline user icon
     * @exception  UnknownHostException  Description of the Exception
     * @exception  IOException           Description of the Exception
     */
    public HLClient(String server,
        int port,
        String login,
        String password,
        String nick,
        short icon) throws UnknownHostException, IOException {

        this(InetAddress.getByName(server),
            port,
            login,
            password,
            nick,
            icon);
    }

    /**
     *  Prepare a connection to the specified host at the specified port with
     *  specified user credentials. Usually "guest" and "" (empty string) will
     *  suffice as credentials.
     *
     * @param  l                         the hotline user login
     * @param  n                         the hotline user nickname
     * @param  i                         the hotline user icon
     * @param  h                         Description of the Parameter
     * @param  pr                        Description of the Parameter
     * @param  ps                        Description of the Parameter
     * @exception  UnknownHostException  Description of the Exception
     * @exception  IOException           Description of the Exception
     */
    public HLClient(InetAddress h,
        int pr,
        String l,
        String ps,
        String n,
        short i) throws UnknownHostException, IOException {

        host = h;
        port = pr;
        login = l;
        password = ps;
        nick = n;
        icon = i;
        tasks = new HashMap();
        transfers = new HashMap();
        hle = new Vector();
        // 2002.12.23 MackyBee modified. 
        // The encoding is set in a setter method. 
        // 2002.12.21 MackyBee modified. 
        //hlp = new HLProtocol(encoding);
        hlp = new HLProtocol();
        // 2002.12.22 MackyBee modified. 
        // HLClientDispatcher refers HLClient's encoding. 
        //hlr = new HLClientDispatcher(this, encoding);
        hlr = new HLClientDispatcher(this);
    }

    /**
     *  Establishes a connection to a Hotline server. This entails opening a
     *  communications socket and dispatching two threads to take care of
     *  reading / writing data to the server. This method is not thread-safe.
     *
     * @exception  IOException     Description of the Exception
     * @throws  BadMagicException  if the server does not respond with the
     *      expected magic value.
     */
    public void connect() throws IOException, BadMagicException {

        if (!connected) {

            connected = true;

            DebuggerOutput.debug("HLClient.connect: opening socket ...");
            socket = new Socket();
            socket.connect(new InetSocketAddress(host, port), 5000);
//            socket = new Socket(host, (char) port);
            DebuggerOutput.debug("HLClient.connect: socket opened");

            // attempt, modify 2002.02.23 by MackyBee
            // I can't understand intention of this process.
            // Then disable these temporary.
            //    socket.setSoTimeout(1000);
            //    socket.setSoLinger(true, 5);

            // attempt, add 2002.02.23 by MackyBee
            // The stream treats small data.
            // So I attempt to add this method setTcpNoDelay(true).
            socket.setTcpNoDelay(true);

            /* 2002.04.29 MackyBee modified */
            output = new DataOutputStream(new BufferedOutputStream(
                socket.getOutputStream()));
            //output = new DataOutputStream(socket.getOutputStream());
            
            /* 2002.05.11 modified to do not use BufferedInputStream, 
               because it has a bug (ID: 4479751).
               2002.04.29 MackyBee experimentally modified */
            input = new DataInputStream(socket.getInputStream());
            
            //input = new DataInputStream(new BufferedInputStream(
            //    socket.getInputStream()));
            
            //input = new DataInputStream(new InterruptableInputStream(
            //    socket.getInputStream()));

            //output.writeBytes(HLProtocol.HTLC_MAGIC);
            // attempt, modified 2002.02.14 by MackyBee
            output.write(HLProtocol.HTLC_MAGIC);

            output.flush();
            DebuggerOutput.debug("HLClient.connect: wrote client magic");

            byte[] buf = new byte[HLProtocol.HTLS_MAGIC_LEN];
            input.readFully(buf, 0, HLProtocol.HTLS_MAGIC_LEN);
            DebuggerOutput.debug("HLClient.connect: read server magic");

            //if(!HLProtocol.HTLS_MAGIC.equals(new String(buf))) {
            // modified 2002.02.14 by MackyBee
            for (int m = 0; m < HLProtocol.HTLS_MAGIC_LEN; m++) {
                if (buf[m] != HLProtocol.HTLS_MAGIC[m]) {
                    socket.close();
                    output = null;
                    input = null;
                    throw new BadMagicException(
                        "not a Hotline server: bad magic");
                }
            }

            reader = new ReaderThread(this, input);
            reader.setBlocked(blocked);
            reader.setName("HLClient ReaderThread " + host.toString());
            reader.start();

            DebuggerOutput.debug("HLClient.connect: (" + getHost().toString() + ").");
        }
    }

    //{{{ unused methods
    /**
     * @since 2002.09.29
     * @author MackyBee
     */
    /*
    protected void setConnected(boolean state) {
        connected = state;
    }
    */

    /**
     * @since 2002.09.29
     * @author MackyBee
     */
    /*
    public boolean isConnected() {
        return (connected);
    }
    */
    //}}}

    /**
     *  Log in with the credentials given during construction or the default
     *  credentials.
     *
     * @exception  IOException           Description of the Exception
     * @exception  HLException           Description of the Exception
     * @exception  InterruptedException  Description of the Exception
     */
    public void login() throws IOException, HLException, InterruptedException {
        login(login, password, nick, icon);
    }

    // Modified on 2002.03.18 by MackyBee
    // int icon -> short icon
    /**
     *  Log in with specified credentials. When the credentials are incorrect,
     *  FailedLoginException is thrown. As a special case of login failure, on >
     *  = 1.5 servers, when one of the registered event listeners rejects the
     *  agreement, an AgreementRejectedException is thrown. For under the 1.5
     *  servers, the agreement is basically ignored; see also {@link
     *  pharerouge.hotline.HLClientListener#handleAgreement}. This method is not
     *  thread-safe.
     *
     * @param  login                     Description of the Parameter
     * @param  password                  Description of the Parameter
     * @param  nick                      Description of the Parameter
     * @param  icon                      Description of the Parameter
     * @exception  IOException           Description of the Exception
     * @exception  HLException           Description of the Exception
     * @exception  InterruptedException  Description of the Exception
     */
    public void login(String login,
        String password,
        String nick,
        short icon)
         throws IOException, HLException, InterruptedException {

        if (login == null || password == null || nick == null) {
            throw new IllegalArgumentException("null argument");
        }

        this.login = login;
        this.password = password;
        this.nick = nick;
        this.icon = icon;

        /*
			Temporary variable set 2002.02.25 by MackyBee
			Sets client version, -71 means 185 means 1.8.5.
			I'd like to be able to users can set a preference value.
			Some servers allows client login permission by the version.
            - 2002.04.14 MackyBee modified the version 1.8.5 > 1.2.3
		*/
        // 2002.11.25 MackyBee EXPERIMENTALLY modified this value to -106 
        // means version 1.5.0.
        byte[] cver =
            {0, -106};
            //{0, 123};
        //{0, -71};

        // 2002.10.26 MackyBee added.
        int numberOfComponents = 3;
        int increment = 0;
        boolean loginExists = false;
        boolean passwordExists = false;
        if (login.equals("")) {
            numberOfComponents--;
        } else {
            loginExists = true;
        }
        if (password.equals("")) {
            numberOfComponents--;
        } else {
            passwordExists = true;
        }

        HLProtocol.DataComponent[] dataComponents = 
            new HLProtocol.DataComponent[numberOfComponents];

        // Note: "No login, but password" is not allowed.
        if (loginExists) {
            dataComponents[0] = hlp.new DataComponent(
                HLProtocol.HTLC_DATA_LOGIN, 
                // 2002.12.03 MackyBee modified. 
                HLProtocol.invert(login.getBytes(encoding)));
                //HLProtocol.invert(EncodingUtil.strictGetBytes(login)));
            increment++;
            if (passwordExists) {
                dataComponents[1] = hlp.new DataComponent(
                    HLProtocol.HTLC_DATA_PASSWORD, 
                    // 2002.12.03 MackyBee modified. 
                    HLProtocol.invert(password.getBytes(encoding)));
                    //HLProtocol.invert(EncodingUtil.strictGetBytes(password)));
                increment++;
            }
            // 2002.12.13 TEMPORARY DEBUG: 
            //  System.out.println("Login: ");
            //  byte[] lbytes = login.getBytes(encoding);
            //  for (int b = 0; b < lbytes.length; b++) {
            //  System.out.print(lbytes[b] + " ");
            //  }
        }

        dataComponents[increment] = hlp.new DataComponent(
            HLProtocol.HTLC_DATA_CLIENT_VERSION, cver);

        // {{{ deactivated old process
        // 2002.10.26 MackyBee replaced this process to other.
        // Modified 2002.02.24 by MackyBee
        // I corrected the login procedure
        /*
        HLProtocol.DataComponent[] dataComponents =
            new HLProtocol.DataComponent[]{
            hlp.new DataComponent(
                HLProtocol.HTLC_DATA_LOGIN, 
                // 2002.05.02 MackyBee modified.
                HLProtocol.invert(strictGetBytes(login))),
                //HLProtocol.invert(login.getBytes())),
            hlp.new DataComponent(
                HLProtocol.HTLC_DATA_PASSWORD, 
                // 2002.05.02 MackyBee modified.
                HLProtocol.invert(strictGetBytes(password))),
                //HLProtocol.invert(password.getBytes())),
            // Added 2002.02.25 by MackyBee.
            hlp.new DataComponent(
                HLProtocol.HTLC_DATA_CLIENT_VERSION,
                cver)

        //hlp.new DataComponent(HLProtocol.HTLC_DATA_NICK, nick.getBytes()),
        //hlp.new DataComponent(HLProtocol.HTLC_DATA_ICON, ToArrayConverters.
        //intToByteArray(icon)),

            };
        */
        // }}}

        HLProtocol.Packet packet =
            hlp.new Packet(HLProtocol.HTLC_HDR_LOGIN,
            nextTrans(),
            dataComponents);

        HLTask task = createTask(packet, false);

        /* Modified 2002.02.24 by MackyBee */
        //DebuggerOutput.debug("HLClient: writing login header");
        DebuggerOutput.debug("HLClient: writing login informations");

        packet.write(output);

        /* Now wait for the server reply. */
        /* Experimentally modify since 2002.02.25 by MackyBee */
        DebuggerOutput.debug("HLClient: waiting for reply from login header");

        // Note: This process also sets serverVersion value.
        try {
            task.getData();
        } finally {
            disposeTask(packet.header.trans);
        }

        /* If we did not get an error, look at the server version. For
           > 1.5 servers, we need to wait until we receive the
           agreement. */
        // 2002.09.29 MackyBee note: 
        // So additionally, under 1.5 servers, we do NOT need to wait until 
        // we receive the agreement.

        // 2002.09.29 MackyBee re-activated.
        // Modified on 2002.02.28 by MackyBee
        if(serverVersion >= 150) {
            try {
                hlr.waitForAgreement();
                if (hlr.agreementAccepted) {

                    // 2002.09.29 MackyBee added.
                    loginSucceeded = true;

                    DebuggerOutput.debug(
                        "HLClient.login: sending login user info...");
                    //sendAgree();

                    // 2002.09.29 MackyBee modified for the sequence change.
                    // Modifed, Added 2002.02.25 by MackyBee
                    //if (serverVersion >= 150) {
                        sendLoginUserInfo(
                            HLProtocol.HTLC_HDR_SEND_USERINFO);

                    //} else {
                    //    sendLoginUserInfo(
                    //        HLProtocol.HTLC_HDR_SEND_OLD_USERINFO);
                    //}

                    // 2002.09.29 added.
//                    preparePeriodicalPacketSender();
                } else {
                    throw new AgreementRejectedException("agreement rejected");
                }
            } catch (InterruptedException e) {
                throw new HLException("wait for agreement interrupted");
            }
        } else {
            // 2002.09.29 MackyBee added.
            loginSucceeded = true;
            // 2002.09.29 MackyBee added this block.
            sendLoginUserInfo(HLProtocol.HTLC_HDR_SEND_OLD_USERINFO);
            // 2002.09.29 added.
//            preparePeriodicalPacketSender();
        }

        //{{{ deactivated old process
        // 2002.09.29 MackyBee note: need thinking. This has moved.
        /* 2002.03.25 MackyBee added experimentally */
        //loginSucceeded = true;

        // 2002.09.29 replaced this to a method.
        // 2002.09.05
        // MackyBee added try-catch statement to deal with future class 
        // relation change.
        /* 2002.03.22 MackyBee modified */
        /*
        try {
            boolean ppsState =
                pharerouge.client.Main.rlo.getBooleanProperty(
                "Toggle.KeepActive");
            if (ppsState) {
                switchPeriodicalPacketSender(ppsState);
            }
        } catch (Exception e) {}
        */
        //}}}
    }

    // 2002.12.22 MackyBee EXPERIMENTALLY added. 
    public HLProtocol getHLProtocol() {
        return (hlp);
    }
         
    // 2002.09.29 added experimentally
//    private void preparePeriodicalPacketSender() {
//        try {
//            boolean ppsState =
//                pharerouge.client.Main.rlo.getBooleanProperty(
//                "Toggle.KeepActive");
//            if (ppsState) {
//                switchPeriodicalPacketSender(ppsState);
//            }
//        } catch (Exception e) {}
//    }

    /* 2002.03.21 MackyBee added experimentally */
    /**
     *  Description of the Method
     *
     * @param  state  Description of the Parameter
     */
    public void switchPeriodicalPacketSender(boolean state) {
        if (state) {
            /* ON */
            if (hlp.pps == null) {
                hlp.pps = new PeriodicalPacketSender(this,
                    HLProtocol.HTLC_HDR_USER_LIST, null, 180000L);
                hlp.pps.start();
                DebuggerOutput.debug("HLClient: The pps was created.");
            }
        } else {
            /* OFF */
            if (hlp.pps != null) {
                hlp.pps.destroy();
                /* 2002.03.22 MackyBee added */
                hlp.pps = null;
                DebuggerOutput.debug("HLClient: The pps was removed.");
            }
        }
    }

    /**
     *  Close the connection with the server. This method must always be called
     *  in order to cleanly terminate running threads. In particular, it must be
     *  called from the {@link HLClientListener#handleDisconnect} method. This
     *  method first closes the socket to the server. Then, all pending tasks
     *  are cancelled, which causes any threads waiting on results from the
     *  getXXX() and waitFor() methods to get an exception.
     *
     * @exception  IOException  Description of the Exception
     */
    public void close() throws IOException {

        setBlocked(false);

        /* Interrupt the reader to let it know that it is being
           closed. */
        if (reader != null) {
            reader.interrupt();
            // 2002.10.14 MackyBee added EXPERIMENTALLY.
            reader = null;
        }

        // 2002.10.14 MackyBee added.
        switchPeriodicalPacketSender(false);

        /* Close the socket to wake up the reader thread. */
        socketClose();

        /* Wait for the reader thread to die. */
        if (reader != null) {
            try {
                DebuggerOutput.debug("HLClient.close: joining reader");
                reader.join();
            } catch (InterruptedException e) {}
        }

        internalDisconnect();

        DebuggerOutput.debug("HLClient.close: (" + getHost().toString() + ").");

    }

    /**
     *  Adds an event handler.
     *
     * @param  h  the HLClientListener object to add
     */
    // 2003.02.11 MackyBee modified. 
    public void addHLClientListener(HLClientListener h) {
        hle.addElement(h);
    }

    /**
     *  Removes an event handler.
     *
     * @param  h  the HLClientListener object to remove
     */
    // 2003.02.11 MackyBee modified. 
    public void removeHLClientListener(HLClientListener h) {
        hle.removeElement(h);
    }

    /**
     *  Returns the server version.
     *
     * @return    the server version (e.g. 100, 150, ...)
     */
    public int getServerVersion() {
        return serverVersion;
    }
    
    /**
     * Usually, called in ReadThread.
     * @since 2003.01.30
     */
    // 2003.01.30 MackyBee added. 
    public String getServerName() {
        if (serverName != null) {
            return (serverName);
        } else {
            return ("a server");
        }
    }
    
    /**
     * Usually, called in the method Machine.setServerName. 
     */
    // 2003.01.30 MackyBee added. 
    public void setServerName(String severName) {
        this.serverName = serverName;
    }

    /**
     *  Returns the host.
     *
     * @return    an InetAddress object
     */
    public InetAddress getHost() {
        return host;
    }

    /**
     *  Returns the port.
     *
     * @return    the port number
     */
    public int getPort() {
        return port;
    }

    /**
     *  Returns the login for this connection.
     *
     * @return    login name for this connection.
     */
    public String getLogin() {
        return login;
    }

    /**
     *  Returns the password for this connection.
     *
     * @return    password for this connection.
     */
    public String getPassword() {
        return password;
    }

    /**
     *  Returns the nick for this connection.
     *
     * @return    nick for this connection.
     */
    public String getNick() {
        return nick;
    }

    /**
     *  Returns the icon for this connection.
     *
     * @return    icon for this connection.
     */
    public int getIcon() {
        return icon;
    }

    /**
     *  Queries connection state.
     *
     * @return    true if the connection with the server is active.
     */
    // 2002.10.21 MackyBee deactivated synchronized keyword.
    public /*synchronized*/ boolean isConnected() {
        return connected;
    }

    /* 2002.03.25 MackyBee added experimentally */
    /**
     *  Gets the loginSucceeded attribute of the HLClient object
     *
     * @return    The loginSucceeded value
     */
    public boolean isLoginSucceeded() {
        return (loginSucceeded);
    }

    /**
     *  Queries whether this connection is blocked.
     *
     * @return    true or false
     */
    public boolean isBlocked() {
        return blocked;
    }

    /**
     *  This method controls whether event processing is active or not. If the
     *  argument is true, then processing of incoming server events is
     *  suspended. This is useful in the case where you need to add a new event
     *  listener, and need to be sure that it does not "miss" any events. It is
     *  OK to call this method at any time, except during {@link #connect}.
     *
     * @param  b  true to block, false to unblock
     */
    public void setBlocked(boolean b) {

        blocked = b;

        if (reader != null) {
            reader.setBlocked(b);
        }
    }

    /* Server commands */
    /**
     *  Requests file (or folder) info.
     *
     * @param  path             the file to get info on
     * @return                  transaction id
     * @exception  IOException  Description of the Exception
     */
    public int requestFileInfo(String path) throws IOException {
        return requestFileInfo(path, true);
    }

    /**
     *  Description of the Method
     *
     * @param  path                      Description of the Parameter
     * @param  disposeWhenReplyReceived  Description of the Parameter
     * @return                           Description of the Return Value
     * @exception  IOException           Description of the Exception
     */
    public int requestFileInfo(String path,
        boolean disposeWhenReplyReceived) throws IOException {

        if (path == null) {
            throw new IllegalArgumentException("path == null");
        }

        int trans = nextTrans();

        HLProtocol.DataComponent[] dataComponents =
            // 2002.12.22 MackyBee modified. 
            hlp.createPathComponents(
            //ComponentFactory.createPathComponents(
            path,
            HLProtocol.HTLC_DATA_DIR,
            true);

        HLProtocol.Packet packet =
            hlp.new Packet(HLProtocol.HTLC_HDR_FILE_GETINFO,
            trans,
            dataComponents);

        createTask(packet, disposeWhenReplyReceived);
        packet.write(output);

        return packet.header.trans;
    }

    /**
     *  Requests the information for the specified account.
     *
     * @param  login            the account to request information for
     * @return                  transaction id
     * @exception  IOException  Description of the Exception
     */
    public int requestAccountInfo(String login) throws IOException {
        return requestAccountInfo(login, true);
    }

    /**
     *  Description of the Method
     *
     * @param  login                     Description of the Parameter
     * @param  disposeWhenReplyReceived  Description of the Parameter
     * @return                           Description of the Return Value
     * @exception  IOException           Description of the Exception
     */
    public int requestAccountInfo(String login,
        boolean disposeWhenReplyReceived) throws IOException {

        if (login == null) {
            throw new IllegalArgumentException("login == null");
        }

        HLProtocol.DataComponent[] dataComponents =
            new HLProtocol.DataComponent[]{
                /* 2002.05.02 MackyBee modified */
                hlp.new DataComponent(HLProtocol.HTLC_DATA_LOGIN, 
                    // 2002.12.03 MackyBee modified. 
                    login.getBytes(encoding))
                    //EncodingUtil.strictGetBytes(login))
                //hlp.new DataComponent(HLProtocol.HTLC_DATA_LOGIN, 
                //login.getBytes()),
            };

        HLProtocol.Packet packet =
            hlp.new Packet(HLProtocol.HTLC_HDR_ACCOUNT_READ,
            nextTrans(),
            dataComponents);

        createTask(packet, disposeWhenReplyReceived);
        packet.write(output);

        return packet.header.trans;
    }

    /**
     *  Upload a file. To monitor the upload progress, you must create a {@link
     *  pharerouge.utils.Meter} object and supply it as the third argument. <p>
     *
     *  The resume flag indicates whether or not this upload should resume an
     *  earlier, partially completed upload. However, this is only possible if
     *  the file already partially exists on the server. You will get a task
     *  error if you try to resume a file that does not already exist.<p>
     *
     *  The proper course of action for an application is to first check (using
     *  {@link #getFileList} for example) whether the file exists on the server,
     *  then to set the resume flag appropriately.
     *
     * @param  local                      the local file
     * @param  remotePath                 the destination path on the server
     * @param  meter                      the non-null Meter object that
     *      monitors this transfer
     * @param  resume                     if true, resumes partially completed
     *      transfer
     * @return                            transaction id
     * @exception  IOException            Description of the Exception
     * @throws  IllegalArgumentException  when meter object is null.
     */
//    public int requestFileUpload(MacFile local,
//        String remotePath,
//        Meter meter,
//        boolean resume) throws IOException {
//
//        return requestFileUpload(local, remotePath, meter, resume, true);
//    }

    /**
     *  Description of the Method
     *
     * @param  local                     Description of the Parameter
     * @param  remotePath                Description of the Parameter
     * @param  meter                     Description of the Parameter
     * @param  resume                    Description of the Parameter
     * @param  disposeWhenReplyReceived  Description of the Parameter
     * @return                           Description of the Return Value
     * @exception  IOException           Description of the Exception
     */
//    public int requestFileUpload(MacFile local,
//        String remotePath,
//        Meter meter,
//        boolean resume,
//        boolean disposeWhenReplyReceived) throws IOException {
//
//        if (meter == null) {
//            throw new IllegalArgumentException("meter object cannot be null");
//        }
//
//        if (remotePath == null || local == null) {
//            throw new IllegalArgumentException("null argument");
//        }
//
//        remotePath = remotePath + HLProtocol.DIR_SEPARATOR +
//            local.getFile().getName();
//
//        HLProtocol.DataComponent[] pathComponents =
//            // 2002.12.22 MackyBee modified. 
//            hlp.createPathComponents(
//            //ComponentFactory.createPathComponents(
//            remotePath,
//            HLProtocol.HTLC_DATA_DIR,
//            true);
//
//        /* Create and serialize this object because we need to know
//           it's length in advance -- apparently Hotline expects the
//           length of download reply in it's entirety (excluding the
//           HTXF block). Changes here mean changes in
//           HLClientDispatcher.java. (There is also a race here; need
//           to add the fti to the tx Transfer object and reuse in
//           HLClientDispatcher). */
//        HLProtocol.FileTransferInfo fti =
//            hlp.new FileTransferInfo(local.getFile().getName(),
//            local.getType(),
//            local.getCreator(),
//            local.getComment(),
//            local.getCreationDate(),
//            local.getModificationDate(),
//            local.getFinderFlags());
//
//        HLProtocol.DataComponent[] dataComponents =
//            new HLProtocol.DataComponent[resume ? 4 : 3];
//
//        // 2002.11.24 MackyBee added this size measuring process.
//        int tdl = 0; // Total Data Length
//        long tempSize = 0;
//        
//        tdl += fti.data.length;
//        tempSize = local.getResourceSize();
//        if (tempSize > 0) {
//            tdl += (int) tempSize;
//            tdl += 16;
//        }
//        tempSize = local.getDataSize();
//        if (tempSize > 0) {
//            tdl += (int) tempSize;
//            tdl += 16;
//        }
//            
//        // 2002.07.26 MackyBee modified. 
//        dataComponents[0] = pathComponents[1];
//        dataComponents[1] = pathComponents[0];
//        dataComponents[2] = hlp.new DataComponent(
//            HLProtocol.HTLC_DATA_XFERSIZE, 
//                ToArrayConverters.intToByteArray(
//                // 2002.11.24 MackyBee modified.
//                tdl));
//                /*
//                16 // DATA block length 
//                + 16 // MACR block length 
//                + fti.data.length 
//                + (int) local.getResourceSize() 
//                + (int) local.getDataSize()));
//                */
//        /*
//        dataComponents[0] = hlp.new DataComponent(
//            HLProtocol.HTLC_DATA_XFERSIZE, ToArrayConverters.intToByteArray(
//            16 // DATA block length 
//            + 16 // MACR block length 
//            + fti.data.length 
//            + (int) local.getResourceSize() 
//            + (int) local.getDataSize()));
//        dataComponents[1] = pathComponents[1];
//        dataComponents[2] = pathComponents[0];
//        */
//
//        if (resume) {
//            dataComponents[3] = hlp.new DataComponent(
//            HLProtocol.HTLC_DATA_RESUME, ToArrayConverters.
//            shortToByteArray(1));
//        }
//
//        HLProtocol.Packet packet =
//            hlp.new Packet(HLProtocol.HTLC_HDR_FILE_PUT,
//            nextTrans(),
//            dataComponents);
//
//        Transfer tx = new Transfer();
//        tx.local = local;
//        tx.remote = remotePath;
//        tx.local_data_size = local.getDataSize();
//        tx.local_rsrc_size = local.getResourceSize();
//        tx.remote_data_size = tx.remote_rsrc_size = 0;
//        tx.meter = meter;
//        
//        createTask(packet, disposeWhenReplyReceived);
//        createTransfer(packet, tx);
//        packet.write(output);
//
//        return packet.header.trans;
//    }
//
//    /**
//     *  Grab the data fork of a file. The meter and inputPipe arguments may not
//     *  be null.
//     *
//     * @param  path             Description of the Parameter
//     * @param  inputPipe        Description of the Parameter
//     * @param  meter            Description of the Parameter
//     * @return                  Description of the Return Value
//     * @exception  IOException  Description of the Exception
//     */
//    public int requestDataForkDownload(String path,
//        PipedInputStream inputPipe,
//        Meter meter) throws IOException {
//
//        if (inputPipe == null) {
//            throw new IllegalArgumentException("inputPipe == null");
//        }
//
//        return requestForkOrFileDownload(path,
//            null,
//            inputPipe,
//            meter,
//            HLProtocol.DL_DATA_FORK,
//            true);
//    }
//
//
//    /**
//     *  Download a file. To monitor the download progress, you must create a
//     *  {@link pharerouge.utils.Meter} object and supply it as the third argument.
//     *  <p>
//     *
//     *  If the local file already exists, then <TT>requestFile</TT> assumes it
//     *  should resume the transfer.
//     *
//     * @param  path                       the file to download
//     * @param  local                      the local file to retrieve to
//     * @param  meter                      the (non-null) Meter object that
//     *      meters this transfer
//     * @return                            transaction id
//     * @exception  IOException            Description of the Exception
//     * @throws  IllegalArgumentException  when meter object is null.
//     */
//    public int requestFileDownload(String path,
//        MacFile local,
//        Meter meter) throws IOException {
//
//        return requestFileDownload(path, local, meter, true);
//    }
//
//    /**
//     *  Description of the Method
//     *
//     * @param  path                      Description of the Parameter
//     * @param  local                     Description of the Parameter
//     * @param  meter                     Description of the Parameter
//     * @param  disposeWhenReplyReceived  Description of the Parameter
//     * @return                           Description of the Return Value
//     * @exception  IOException           Description of the Exception
//     */
//    public int requestFileDownload(String path,
//        MacFile local,
//        Meter meter,
//        boolean disposeWhenReplyReceived) throws IOException {
//
//        if (local == null) {
//            throw new IllegalArgumentException("local == null");
//        }
//
//        return requestForkOrFileDownload(path,
//            local,
//            null,
//            meter,
//            HLProtocol.DL_WHOLE_FILE,
//            disposeWhenReplyReceived);
//    }
//
//    /**
//     *  Requests the file list.
//     *
//     * @param  path             the path whose contents to retrieve
//     * @return                  transaction id
//     * @exception  IOException  Description of the Exception
//     */
//    public int requestFileList(String path) throws IOException {
//        return requestFileList(path, true);
//    }
//
//    /**
//     *  Description of the Method
//     *
//     * @param  path                      Description of the Parameter
//     * @param  disposeWhenReplyReceived  Description of the Parameter
//     * @return                           Description of the Return Value
//     * @exception  IOException           Description of the Exception
//     */
//    public int requestFileList(String path,
//        boolean disposeWhenReplyReceived) throws IOException {
//
//        if (path == null) {
//            throw new IllegalArgumentException("path == null");
//        }
//
//        HLProtocol.DataComponent[] pathComponents =
//            // 2002.12.22 MackyBee modified. 
//            hlp.createPathComponents(
//            //ComponentFactory.createPathComponents(
//            path,
//            HLProtocol.HTLC_DATA_DIR,
//            false);
//
//        HLProtocol.DataComponent[] dataComponents =
//            new HLProtocol.DataComponent[]{
//            pathComponents[0],
//            };
//
//        HLProtocol.Packet packet =
//            hlp.new Packet(HLProtocol.HTLC_HDR_DIR_LIST,
//            nextTrans(),
//            dataComponents);
//
//        createTask(packet, disposeWhenReplyReceived);
//        packet.write(output);
//
//        return packet.header.trans;
//    }
//
//    /* 2002.05.28 MackyBee experimentally added */
//    public HLProtocol.NewsBundleListComponent[] getNewsBundleList() 
//        throws IOException, HLException, InterruptedException {
//
//        HLProtocol.Packet packet =
//            hlp.new Packet(HLProtocol.HTLC_HDR_GET_NEWS_BUNDLE,
//            nextTrans(),
//            null);
//        /* 2002.05.31 modified */
//        createTask(packet, false);
//        //createTask(packet, true);
//        packet.write(output);
//        
//        Object o = waitFor(packet.header.trans);
//        
//        return (HLProtocol.NewsBundleListComponent[])o;
//    }
//
//    /* 2002.06.02 MackyBee experimentally added */
//    public HLProtocol.NewsBundleListComponent[] getNewsCategoryList(
//        byte[] bundleNameBytes) throws IOException, HLException, 
//        InterruptedException {
//        
//        /* replaced following process this */
//        HLProtocol.NewsBundleListComponent[] nbcs = {
//            hlp.new NewsBundleListComponent(bundleNameBytes)};
//        /*
//        byte[] dcData = new byte[bundleNameBytes.length + 1];
//        dcData[0] = (byte) bundleNameBytes.length;
//        System.arraycopy(bundleNameBytes, 0, dcData, 1, dcData.length - 1);
//        
//        HLProtocol.DataComponent[] dcs = new HLProtocol.DataComponent[1];
//        dcs[0] = hlp.new DataComponent(
//            HLProtocol.HTLC_DATA_NEWS_COMPONENT, 
//            dcData);
//        */
//        
//        HLProtocol.Packet packet =
//            hlp.new Packet(HLProtocol.HTLC_HDR_GET_NEWS_BUNDLE, 
//            nextTrans(), 
//            nbcs);
//            //dcs);
//        
//        createTask(packet, false);
//        packet.write(output);
//        Object o = waitFor(packet.header.trans);
//        
//        return (HLProtocol.NewsBundleListComponent[])o;
//    }
//
//    /* 2002.06.10 MackyBee added */
//    public HLProtocol.NewsBundleListComponent[] getNewsCategoryList(
//        NewsComponent[] ncs) throws IOException, HLException, 
//        InterruptedException {
//
//        /* Delegating */
//        HLProtocol.NewsThreadListComponent[] ntlc = {
//            hlp.new NewsThreadListComponent(ncs)};
//
//        HLProtocol.Packet packet =
//            hlp.new Packet(HLProtocol.HTLC_HDR_GET_NEWS_BUNDLE, 
//            nextTrans(), 
//            ntlc);
//
//        createTask(packet, false);
//        packet.write(output);
//        Object o = waitFor(packet.header.trans);
//        
//        return ((HLProtocol.NewsBundleListComponent[]) o);
//    }
    
    /* Disabled. 2002.06.04 MackyBee added.
       Note: To return non-array object is correct. */
    /*
    public HLProtocol.NewsThreadListComponent getNewsThreadList(
        byte[] bundle, byte[] category) 
        throws IOException, HLException, InterruptedException {

        HLProtocol.NewsThreadListComponent[] ntlc = {
            hlp.new NewsThreadListComponent(bundle, category)};

        HLProtocol.Packet packet =
            hlp.new Packet(HLProtocol.HTLC_HDR_GET_NEWS_THREADS, 
            nextTrans(), 
            ntlc);

        createTask(packet, false);
        packet.write(output);
        Object o = waitFor(packet.header.trans);

        return ((HLProtocol.NewsThreadListComponent) o);
    }
    */

    /* 2002.06.10 MackyBee added */
//    public HLProtocol.NewsThreadListComponent getNewsThreadList(
//        NewsComponent[] ncs) 
//        throws IOException, HLException, InterruptedException {
//
//        /* Note: Delegating */
//        HLProtocol.NewsThreadListComponent[] ntlc = {
//            hlp.new NewsThreadListComponent(ncs)};
//
//        HLProtocol.Packet packet =
//            hlp.new Packet(HLProtocol.HTLC_HDR_GET_NEWS_THREADS, 
//            nextTrans(), 
//            ntlc);
//
//        createTask(packet, false);
//        packet.write(output);
//        Object o = waitFor(packet.header.trans);
//
//        return ((HLProtocol.NewsThreadListComponent) o);
//    }
//
//    /* Disabled.
//       2002.06.07 MackyBee modified the return type. 
//       2002.06.04 MackyBee added */
//    /*public String getNewsArticle(
//    //public HLProtocol.NewsArticleComponent getNewsArticle(
//        byte[] bundle, byte[] category, int index) 
//        throws IOException, HLException, InterruptedException {
//            */
//        /* Note: composite DataComponent. */
//        /*HLProtocol.DataComponent[] dc = {
//            hlp.new NewsThreadListComponent(bundle, category), 
//            hlp.new NewsArticleIndexComponent(index), */
//            /* Note: Provisionally, hard coded. */
//            /*hlp.new NewsArticleMIMETypeComponent("text/plain")};
//
//        HLProtocol.Packet packet =
//            hlp.new Packet(HLProtocol.HTLC_HDR_GET_NEWS_ARTICLE, 
//            nextTrans(), 
//            dc);
//
//        createTask(packet, false);
//        packet.write(output);
//        Object o = waitFor(packet.header.trans);
//        */
//        /* 2002.06.07 MackyBee modified */
//        /*return (o.toString());
//        //return ((HLProtocol.NewsArticleComponent) o);
//    }*/
//
//    /* 2002.06.10 MackyBee added */
//    public String getNewsArticle(NewsComponent[] ncs, int index) 
//        throws IOException, HLException, InterruptedException {
//
//        /* Note: composite DataComponent. */
//        HLProtocol.DataComponent[] dc = {
//            hlp.new NewsThreadListComponent(ncs), 
//            hlp.new NewsArticleIndexComponent(index), 
//            /* Note: Provisionally, hard coded. */
//            hlp.new NewsArticleMIMETypeComponent("text/plain")};
//
//        HLProtocol.Packet packet =
//            hlp.new Packet(HLProtocol.HTLC_HDR_GET_NEWS_ARTICLE, 
//            nextTrans(), 
//            dc);
//
//        createTask(packet, false);
//        packet.write(output);
//        Object o = waitFor(packet.header.trans);
//
//        /* 2002.06.07 MackyBee modified */
//        return (o.toString());
//        //return ((HLProtocol.NewsArticleComponent) o);
//    }
//
//    /* 2002.06.15 MackyBee added. */
//    public int requestNewsBundleCreate(byte[] title, NewsComponent[] ncs) 
//        throws IOException, HLException, InterruptedException {
//
//        HLProtocol.DataComponent[] dc = {
//            hlp.new DataComponent(HLProtocol.HTLC_DATA_NEWS_BUNDLE_TITLE, 
//                title),
//            hlp.new NewsThreadListComponent(ncs) 
//            };
//
//        HLProtocol.Packet packet =
//            hlp.new Packet(HLProtocol.HTLC_HDR_CREATE_NEWS_BUNDLE, 
//            nextTrans(), 
//            dc);
//
//        createTask(packet, true);
//        packet.write(output);
//        
//        return (packet.header.trans);
//    }
//
//    /* 2002.06.15 MackyBee added. */
//    public int requestNewsCategoryCreate(byte[] title, NewsComponent[] ncs) 
//        throws IOException, HLException, InterruptedException {
//
//        HLProtocol.DataComponent[] dc = {
//            hlp.new DataComponent(HLProtocol.HTLC_DATA_NEWS_CATEGORY_TITLE, 
//                title),
//            hlp.new NewsThreadListComponent(ncs) 
//            };
//
//        HLProtocol.Packet packet =
//            hlp.new Packet(HLProtocol.HTLC_HDR_CREATE_NEWS_CATEGORY, 
//            nextTrans(), 
//            dc);
//
//        createTask(packet, true);
//        packet.write(output);
//        
//        return (packet.header.trans);
//    }
//
//    /* 2002.06.15 MackyBee added. */
//    /** Corresponding to multi category, multi threaded news. */
//    public int requestNewsArticlePost(NewsComponent[] ncs, 
//        int targetID, byte[] title, 
//        byte[] article, byte[] mimeType) 
//        throws IOException, HLException, InterruptedException {
//
//        HLProtocol.DataComponent[] dc = {
//            hlp.new NewsThreadListComponent(ncs), 
//            hlp.new NewsArticleIndexComponent(targetID), 
//            hlp.new DataComponent(HLProtocol.HTLC_DATA_NEWS_ARTICLE_TITLE, 
//                title), 
//            hlp.new DataComponent(HLProtocol.HTLC_DATA_NEWS_UNKNOWN_01), 
//            hlp.new DataComponent(HLProtocol.HTLC_DATA_NEWS_ARTICLE_MIME_TYPE,
//                mimeType), 
//            hlp.new DataComponent(HLProtocol.HTLS_DATA_NEWS_ARTICLE, 
//                article)
//            };
//
//        HLProtocol.Packet packet =
//            hlp.new Packet(HLProtocol.HTLC_HDR_POST_NEWS_ARTICLE, 
//            nextTrans(), 
//            dc);
//
//        createTask(packet, true);
//        packet.write(output);
//        
//        return (packet.header.trans);
//    }
//
//    /* 2002.06.17 MackyBee added. */
//    public int requestNewsArticleDelete(NewsComponent[] ncs, short index, 
//        short type)
//        throws IOException, HLException, InterruptedException {
//
//        HLProtocol.DataComponent[] dc = {
//            hlp.new NewsThreadListComponent(ncs), 
//            hlp.new DataComponent(HLProtocol.HTLC_DATA_NEWS_ARTICLE_INDEX, 
//                ToArrayConverters.shortToByteArray(index)),
//            hlp.new DataComponent(HLProtocol.HTLC_DATA_NEWS_DELETE_TYPE, 
//                ToArrayConverters.shortToByteArray(type))
//            };
//
//        HLProtocol.Packet packet =
//            hlp.new Packet(HLProtocol.HTLC_HDR_DELETE_NEWS_ARTICLE, 
//            nextTrans(), 
//            dc);
//
//        createTask(packet, true);
//        packet.write(output);
//
//        return (packet.header.trans);
//    }
//
//    /* 2002.06.17 MackyBee added. */
//    /** The "Container" means Bundles and Categories. */
//    public int requestNewsContainerDelete(NewsComponent[] ncs) 
//        throws IOException, HLException, InterruptedException {
//
//        HLProtocol.DataComponent[] dc = {
//            hlp.new NewsThreadListComponent(ncs), 
//            };
//
//        HLProtocol.Packet packet =
//            hlp.new Packet(HLProtocol.HTLC_HDR_DELETE_NEWS_CONTAINER, 
//            nextTrans(), 
//            dc);
//
//        createTask(packet, true);
//        packet.write(output);
//
//        return (packet.header.trans);
//    }

    /**
     *  Requests the user list.
     *
     * @return                  transaction id
     * @exception  IOException  Description of the Exception
     */
    public int requestUserList() throws IOException {

        return requestUserList(true);
    }


    /* 2002.03.22 MackyBee modified */
    /**
     *  Description of the Method
     *
     * @param  disposeWhenReplyReceived  Description of the Parameter
     * @return                           Description of the Return Value
     * @exception  IOException           Description of the Exception
     */
    public int requestUserList(boolean disposeWhenReplyReceived) throws
        IOException {

        HLProtocol.Packet packet =
            hlp.new Packet(HLProtocol.HTLC_HDR_USER_LIST,
            nextTrans(),
            null);

        createTask(packet, disposeWhenReplyReceived);
        packet.write(output);

        return packet.header.trans;
    }


    /**
     *  Requests the news.
     *
     * @return                  transaction id
     * @exception  IOException  Description of the Exception
     */
    public int requestNews() throws IOException {

        return requestNews(true);
    }


    /**
     *  Description of the Method
     *
     * @param  disposeWhenReplyReceived  Description of the Parameter
     * @return                           Description of the Return Value
     * @exception  IOException           Description of the Exception
     */
    public int requestNews(boolean disposeWhenReplyReceived) throws IOException {

        HLProtocol.Packet packet =
            hlp.new Packet(HLProtocol.HTLC_HDR_NEWS_GET,
            nextTrans(),
            null);

        createTask(packet, disposeWhenReplyReceived);
        packet.write(output);

        return packet.header.trans;
    }


    /**
     *  Requests user info.
     *
     * @param  sock             sock of user to get info on
     * @return                  transaction id
     * @exception  IOException  Description of the Exception
     */
    public int requestUserInfo(int sock) throws IOException {

        return requestUserInfo(sock, true);
    }


    /**
     *  Description of the Method
     *
     * @param  sock                      Description of the Parameter
     * @param  disposeWhenReplyReceived  Description of the Parameter
     * @return                           Description of the Return Value
     * @exception  IOException           Description of the Exception
     */
    public int requestUserInfo(int sock,
        boolean disposeWhenReplyReceived) throws IOException {

        HLProtocol.DataComponent[] dataComponents =
            new HLProtocol.DataComponent[]{
            hlp.new DataComponent(HLProtocol.HTLC_DATA_SOCKET,
            ToArrayConverters.intToByteArray(sock)),
            };

        HLProtocol.Packet packet =
            hlp.new Packet(HLProtocol.HTLC_HDR_USER_GETINFO,
            nextTrans(),
            dataComponents);

        createTask(packet, disposeWhenReplyReceived);
        packet.write(output);

        return packet.header.trans;
    }


    /**
     *  Requests account information and returns it; returns null on error. Note
     *  that the password field in the {@link
     *  pharerouge.hotline.HLProtocol.AccountInfo} structure will always contain
     *  garbage. If you wish to modify an existing account but not change the
     *  password, you must supply <TT>null</TT> as the password in {@link
     *  pharerouge.hotline.HLClient#sendAccountChange}. This function blocks until
     *  the requested data has been received.
     *
     * @param  login                     Description of the Parameter
     * @return                           account information
     * @exception  IOException           Description of the Exception
     * @exception  HLException           Description of the Exception
     * @exception  InterruptedException  Description of the Exception
     */
    public HLProtocol.AccountInfo getAccountInfo(String login) 
        throws IOException, HLException, InterruptedException {
        Object o = waitFor(requestAccountInfo(login, false));
        return (HLProtocol.AccountInfo) o;
    }

    /**
     *  Requests info on a file and returns it; returns null on error. This
     *  function blocks until the requested data has been received.
     *
     * @param  path                      the filesystem object to get info on
     * @return                           information on the filesystem object
     * @exception  IOException           Description of the Exception
     * @exception  HLException           Description of the Exception
     * @exception  InterruptedException  Description of the Exception
     */
    public HLProtocol.FileInfo getFileInfo(String path) throws IOException, 
        HLException, InterruptedException {
        Object o = waitFor(requestFileInfo(path, false));
        return (HLProtocol.FileInfo) o;
    }


    /**
     *  Requests the file list and returns it; returns null on error. This
     *  function blocks until the requested data has been received.
     *
     * @param  path                      the path to retrieve
     * @return                           an array of files
     * @exception  IOException           Description of the Exception
     * @exception  HLException           Description of the Exception
     * @exception  InterruptedException  Description of the Exception
     */
//    public HLProtocol.FileListComponent[] getFileList(String path) throws IOException, HLException, InterruptedException {
//
//        Object o = waitFor(requestFileList(path, false));
//        return (HLProtocol.FileListComponent[]) o;
//    }


    /* 2002.03.22 MackyBee modified */
    /**
     *  Requests the user list and returns it; returns null on error. This
     *  function blocks until the requested data has been received.
     *
     * @return                           an array of users
     * @exception  IOException           Description of the Exception
     * @exception  HLException           Description of the Exception
     * @exception  InterruptedException  Description of the Exception
     */
    public HLProtocol.UserListComponent[] getUserList() throws IOException,
        HLException, InterruptedException {

        Object o = waitFor(requestUserList(false));
        return (HLProtocol.UserListComponent[]) o;
    }


    /**
     *  Requests the news and returns it; returns null on error. This function
     *  blocks until the requested data has been received.
     *
     * @return                           the news
     * @exception  IOException           Description of the Exception
     * @exception  HLException           Description of the Exception
     * @exception  InterruptedException  Description of the Exception
     */
    public String getNews() throws IOException, HLException, InterruptedException {

        Object o = waitFor(requestNews(false));
        return (String) o;
    }


    /**
     *  Requests user info and returns it; returns null on error. This function
     *  blocks until the requested data has been received.
     *
     * @param  sock                      Description of the Parameter
     * @return                           user info
     * @exception  IOException           Description of the Exception
     * @exception  HLException           Description of the Exception
     * @exception  InterruptedException  Description of the Exception
     */
    public String getUserInfo(int sock) throws IOException, HLException, InterruptedException {

        Object o = waitFor(requestUserInfo(sock, false));
        return (String) o;
    }


    /**
     *  Changes the name and comment of a file (or directory).
     *
     * @param  path             the file to change
     * @param  newName          the new name of the file
     * @param  newComment       the new comment of the file
     * @return                  Description of the Return Value
     * @exception  IOException  Description of the Exception
     */
    public int requestFileInfoChange(String path,
        String newName,
        String newComment) throws IOException {

        return requestFileInfoChange(path, newName, newComment, true);
    }


    /**
     *  Description of the Method
     *
     * @param  path                      Description of the Parameter
     * @param  newName                   Description of the Parameter
     * @param  newComment                Description of the Parameter
     * @param  disposeWhenReplyReceived  Description of the Parameter
     * @return                           Description of the Return Value
     * @exception  IOException           Description of the Exception
     */
    public int requestFileInfoChange(String path,
        String newName,
        String newComment,
        boolean disposeWhenReplyReceived) throws IOException {

        if (path == null || newName == null || newComment == null) {
            throw new IllegalArgumentException("null argument");
        }

        HLProtocol.DataComponent[] pathComponents =
            // 2002.12.22 MackyBee modified. 
            hlp.createPathComponents(
            //ComponentFactory.createPathComponents(
            path,
            HLProtocol.HTLC_DATA_DIR,
            true);

        HLProtocol.DataComponent[] dataComponents =
            new HLProtocol.DataComponent[]{
                pathComponents[1],
                pathComponents[0],
                /* 2002.05.02 MackyBee modified */
                hlp.new DataComponent(HLProtocol.HTLC_DATA_FILE_RENAME, 
                    // 2002.12.03 MackyBee modified. 
                    newName.getBytes(encoding)),
                    //EncodingUtil.strictGetBytes(newName)),
                hlp.new DataComponent(HLProtocol.HTLS_DATA_FILE_COMMENT, 
                   // 2002.12.03 MackyBee modified. 
                    newComment.getBytes(encoding))
                    //EncodingUtil.strictGetBytes(newComment))
                //hlp.new DataComponent(HLProtocol.HTLC_DATA_FILE_RENAME, 
                //    newName.getBytes()),
                //hlp.new DataComponent(HLProtocol.HTLS_DATA_FILE_COMMENT, 
                //    newComment.getBytes()),
            };

        HLProtocol.Packet packet =
            hlp.new Packet(HLProtocol.HTLC_HDR_FILE_SETINFO,
            nextTrans(),
            dataComponents);

        createTask(packet, disposeWhenReplyReceived);
        packet.write(output);

        return packet.header.trans;
    }

    /**
     *  Creates an alias (like a symbolic link). The below example creates an
     *  alias "testfile" in the root directory, which points to the file
     *  "testfile" in the directory "test":<P>
     *
     *  <TT>hlc.sendFileMakeAlias(":test:testfile", ":");</TT> <BR>
     *
     *
     * @param  source           the object on the server to make an alias to
     * @param  target           the location of the alias
     * @return                  transaction id
     * @exception  IOException  Description of the Exception
     */
    public int requestFileMakeAlias(String source,
        String target) throws IOException {
        return requestFileMakeAlias(source, target, true);
    }

    /**
     *  Description of the Method
     *
     * @param  source                    Description of the Parameter
     * @param  target                    Description of the Parameter
     * @param  disposeWhenReplyReceived  Description of the Parameter
     * @return                           Description of the Return Value
     * @exception  IOException           Description of the Exception
     */
    public int requestFileMakeAlias(String source,
        String target,
        boolean disposeWhenReplyReceived) throws IOException {

        return requestFileMoveOrMakeAlias(HLProtocol.HTLC_HDR_FILE_MAKE_ALIAS,
            source,
            target,
            disposeWhenReplyReceived);
    }

    /**
     *  Moves a file.
     *
     * @param  source           the file on the server to move
     * @param  target           the location to move the file to
     * @return                  transaction id
     * @exception  IOException  Description of the Exception
     */
    public int requestFileMove(String source,
        String target) throws IOException {

        return requestFileMove(source, target, true);
    }

    /**
     *  Description of the Method
     *
     * @param  source                    Description of the Parameter
     * @param  target                    Description of the Parameter
     * @param  disposeWhenReplyReceived  Description of the Parameter
     * @return                           Description of the Return Value
     * @exception  IOException           Description of the Exception
     */
    public int requestFileMove(String source,
        String target,
        boolean disposeWhenReplyReceived) throws IOException {

        return requestFileMoveOrMakeAlias(HLProtocol.HTLC_HDR_FILE_MOVE,
            source,
            target,
            disposeWhenReplyReceived);
    }

    /**
     *  Deletes a file on the server.
     *
     * @param  path             the file on the server to delete
     * @return                  transaction id
     * @exception  IOException  Description of the Exception
     */
    public int requestFileDelete(String path) throws IOException {
        return requestFileDelete(path, true);
    }

    /**
     *  Description of the Method
     *
     * @param  path                      Description of the Parameter
     * @param  disposeWhenReplyReceived  Description of the Parameter
     * @return                           Description of the Return Value
     * @exception  IOException           Description of the Exception
     */
    public int requestFileDelete(String path,
        boolean disposeWhenReplyReceived) throws IOException {

        HLProtocol.DataComponent[] pathComponents =
            // 2002.12.22 MackyBee modified. 
            hlp.createPathComponents(
            //ComponentFactory.createPathComponents(
            path,
            HLProtocol.HTLC_DATA_DIR,
            true);

        HLProtocol.DataComponent[] dataComponents =
            new HLProtocol.DataComponent[]{
            pathComponents[1],
            pathComponents[0],
            };

        HLProtocol.Packet packet =
            hlp.new Packet(HLProtocol.HTLC_HDR_FILE_DELETE,
            nextTrans(),
            dataComponents);

        createTask(packet, disposeWhenReplyReceived);
        packet.write(output);

        return packet.header.trans;
    }

    /**
     *  Creates a directory.
     *
     * @param  path             the directory to create on the server
     * @return                  transaction id
     * @exception  IOException  Description of the Exception
     */
    public int requestDirectoryCreate(String path) throws IOException {
        return requestDirectoryCreate(path, true);
    }

    /**
     *  Description of the Method
     *
     * @param  path                      Description of the Parameter
     * @param  disposeWhenReplyReceived  Description of the Parameter
     * @return                           Description of the Return Value
     * @exception  IOException           Description of the Exception
     */
    public int requestDirectoryCreate(String path,
        boolean disposeWhenReplyReceived) throws IOException {
            
        HLProtocol.DataComponent[] pathComponents =
            // 2002.12.22 MackyBee modified. 
            hlp.createPathComponents(
            //ComponentFactory.createPathComponents(
            path,
            HLProtocol.HTLC_DATA_DIR,
            true);

        HLProtocol.DataComponent[] dataComponents =
            new HLProtocol.DataComponent[]{
            pathComponents[1],
            pathComponents[0],
            };

        HLProtocol.Packet packet =
            hlp.new Packet(HLProtocol.HTLC_HDR_DIR_CREATE,
            nextTrans(),
            dataComponents);

        createTask(packet, disposeWhenReplyReceived);
        packet.write(output);

        return packet.header.trans;
    }

    // Modified on 2002.03.18 by MackyBee
    // int icon -> short icon
    /**
     *  Changes our nickname and icon. The server may simply ignore us if it
     *  doesn't like the new settings. Otherwise the server sends a "user
     *  change" message. Either way, the nick and icon as returned by {@link
     *  #getNick} and {@link getIcon} are changed by this method.
     *
     * @param  nick             Description of the Parameter
     * @param  icon             Description of the Parameter
     * @exception  IOException  Description of the Exception
     */
    public void sendUserChange(String nick, short icon) throws IOException {
        //public void sendUserChange(String nick,
        //                           int icon) throws IOException {

        if (nick == null) {
            throw new IllegalArgumentException("nick == null)");
        }

        this.nick = nick;
        this.icon = icon;

        HLProtocol.DataComponent[] dataComponents =
            new HLProtocol.DataComponent[]{
        // Modified on 2002.03.18 by MackyBee
        // the 'icon' should be a short value.
            hlp.new DataComponent(HLProtocol.HTLC_DATA_ICON,
            ToArrayConverters.shortToByteArray(icon)),
        //hlp.new DataComponent(HLProtocol.HTLC_DATA_ICON, ToArrayConverters.charToByteArray((char) icon)),
                /* 2002.05.02 MackyBee modified */
                hlp.new DataComponent(HLProtocol.HTLC_DATA_NICK, 
                    // 2002.12.03 MackyBee modified. 
                    nick.getBytes(encoding))
                    //EncodingUtil.strictGetBytes(nick))
            //hlp.new DataComponent(HLProtocol.HTLC_DATA_NICK, nick.getBytes())
            };

        HLProtocol.Packet packet =
            hlp.new Packet(HLProtocol.HTLC_HDR_USER_CHANGE,
            nextTrans(),
            dataComponents);

        packet.write(output);
    }

    /**
     *  Deletes the specified account.
     *
     * @param  login            the account to delete
     * @return                  transaction id
     * @exception  IOException  Description of the Exception
     */
    public int requestAccountDelete(String login) throws IOException {
        return requestAccountDelete(login, true);
    }

    /**
     *  Description of the Method
     *
     * @param  login                     Description of the Parameter
     * @param  disposeWhenReplyReceived  Description of the Parameter
     * @return                           Description of the Return Value
     * @exception  IOException           Description of the Exception
     */
    public int requestAccountDelete(String login,
        boolean disposeWhenReplyReceived) throws IOException {

        if (login == null) {
            throw new IllegalArgumentException("login == null)");
        }

        HLProtocol.DataComponent[] dataComponents =
            new HLProtocol.DataComponent[]{
                /* 2002.05.02 MackyBee modified */
                hlp.new DataComponent(HLProtocol.HTLC_DATA_LOGIN, 
                    // 2002.12.03 MackyBee modified. 
                    HLProtocol.invert(login.getBytes(encoding)))
                    //HLProtocol.invert(EncodingUtil.strictGetBytes(login)))
            //hlp.new DataComponent(HLProtocol.HTLC_DATA_LOGIN, 
            //HLProtocol.invert(login.getBytes())),
            };

        HLProtocol.Packet packet =
            hlp.new Packet(HLProtocol.HTLC_HDR_ACCOUNT_DELETE,
            nextTrans(),
            dataComponents);

        createTask(packet, disposeWhenReplyReceived);
        packet.write(output);

        return packet.header.trans;
    }

    /**
     *  Modifies an existing user account. If password is <TT>null</TT> , then
     *  the existing password is not changed.
     *
     * @param  login            the login name
     * @param  password         the new password
     * @param  nick             the new nickname
     * @param  privileges       the new privileges (see {@link
     *      pharerouge.hotline.HLProtocol}
     * @return                  transaction id
     * @exception  IOException  Description of the Exception
     */
    public int requestAccountModify(String login,
        String password,
        String nick,
        long privileges) throws IOException {

        return requestAccountModify(login, password, nick, privileges, true);
    }

    /**
     *  Description of the Method
     *
     * @param  login                     Description of the Parameter
     * @param  password                  Description of the Parameter
     * @param  nick                      Description of the Parameter
     * @param  privileges                Description of the Parameter
     * @param  disposeWhenReplyReceived  Description of the Parameter
     * @return                           Description of the Return Value
     * @exception  IOException           Description of the Exception
     */
    public int requestAccountModify(String login,
        String password,
        String nick,
        long privileges,
        boolean disposeWhenReplyReceived) throws IOException {

        return requestAccountCreateOrModify(login,
            password,
            nick,
            privileges,
            HLProtocol.HTLC_HDR_ACCOUNT_MODIFY,
            disposeWhenReplyReceived);
    }

    /**
     *  Create a new user account.
     *
     * @param  login            the login name
     * @param  password         the new password
     * @param  nick             the new nickname
     * @param  privileges       the new privileges (see {@link
     *      pharerouge.hotline.HLProtocol.AccountInfo}
     * @return                  transaction id
     * @exception  IOException  Description of the Exception
     */
    public int requestAccountCreate(String login,
        String password,
        String nick,
        long privileges) throws IOException {

        return requestAccountCreate(login, password, nick, privileges, true);
    }

    /**
     *  Description of the Method
     *
     * @param  login                     Description of the Parameter
     * @param  password                  Description of the Parameter
     * @param  nick                      Description of the Parameter
     * @param  privileges                Description of the Parameter
     * @param  disposeWhenReplyReceived  Description of the Parameter
     * @return                           Description of the Return Value
     * @exception  IOException           Description of the Exception
     */
    public int requestAccountCreate(String login,
        String password,
        String nick,
        long privileges,
        boolean disposeWhenReplyReceived) throws IOException {

        return requestAccountCreateOrModify(login,
            password,
            nick,
            privileges,
            HLProtocol.HTLC_HDR_ACCOUNT_CREATE,
            disposeWhenReplyReceived);
    }

    /**
     * Sets a character encoding used when calling method String.getBytes(
     * String).
     * @return true if setting was successful, false otherwise.
     * @since 2002.12.13
     */
    // 2002.12.13 MackyBee added this method. 
    public boolean setEncoding(String newEncoding) {
        if (newEncoding != null) {
            // NOTE: I'd like to add here a checking process 
            // UnsupportedEncodingException possibility.
            encoding = newEncoding;
            // 2002.12.23 MackyBee added. 
            hlp.setEncoding(encoding);
            return (true);
        } else {
            return (false);
        }
    }

    /**
     * @return current character encoding
     * @see setEncoding(String)
     */
    // 2002.12.13 MackyBee added this method. 
    public String getEncoding() {
        return (encoding);
    }
        
    /**
     *  Sends a string to the chat. HL servers do not consider chat to be a
     *  task. hxd on the other hand does consider chat to be a task, and will
     *  send a task error object in case we do not have permission to send that.
     *  What's strange however is that while hxd will send a task error for a
     *  failed chat request, it will not send a task complete object for a
     *  successfull chat request.
     *
     * @param  s                the string to send
     * @exception  IOException  Description of the Exception
     */
    public void sendChat(String s) throws IOException {

        if (s == null) {
            throw new IllegalArgumentException("s == null");
        }

        /* 2002.05.02 MackyBee added temporary */
        //DebuggerOutput.putHexString(new String(
        //    s.getBytes(pharerouge.client.Main.DEFAULT_ENCODING)).getBytes());
        
        HLProtocol.DataComponent[] dataComponents =
            new HLProtocol.DataComponent[] {
                /* 2002.05.02 modified, 
                   2002.05.01 MackyBee experimentally modified */
                hlp.new DataComponent(HLProtocol.HTLC_DATA_CHAT, 
                // 2002.12.03 MackyBee modified. 
                s.getBytes(encoding))
                //EncodingUtil.strictGetBytes(s))
                //hlp.new DataComponent(HLProtocol.HTLC_DATA_CHAT, 
                //  s.getBytes())
            };

        HLProtocol.Packet packet =
            hlp.new Packet(HLProtocol.HTLC_HDR_CHAT,
            nextTrans(),
            dataComponents);

        packet.write(output);
    }

    /**
     *  Sends an action string to the chat.
     *
     * @param  s                the action to send
     * @exception  IOException  Description of the Exception
     */
    public void sendChatAction(String s) throws IOException {

        if (s == null) {
            throw new IllegalArgumentException("s == null");
        }

        HLProtocol.DataComponent[] dataComponents =
            new HLProtocol.DataComponent[]{
                /* 2002.05.02 MackyBee modified */
                hlp.new DataComponent(HLProtocol.HTLC_DATA_CHAT, 
                // 2002.12.03 MackyBee modified. 
                s.getBytes(encoding)),
                //EncodingUtil.strictGetBytes(s)),
                //hlp.new DataComponent(HLProtocol.HTLC_DATA_CHAT, 
                //s.getBytes()),
                hlp.new DataComponent(HLProtocol.HTLC_DATA_OPTION, 
                    ToArrayConverters.shortToByteArray(1)),
            };

        HLProtocol.Packet packet =
            hlp.new Packet(HLProtocol.HTLC_HDR_CHAT,
            nextTrans(),
            dataComponents);

        packet.write(output);
    }

    /**
     *  Create a private chat with another user.
     *
     * @param  sock             the sock of the user to privchat with
     * @return                  transaction id
     * @exception  IOException  Description of the Exception
     */
    public int requestPrivateChatCreate(int sock) throws IOException {
        return requestPrivateChatCreate(sock, true);
    }

    /**
     *  Description of the Method
     *
     * @param  sock                      Description of the Parameter
     * @param  disposeWhenReplyReceived  Description of the Parameter
     * @return                           Description of the Return Value
     * @exception  IOException           Description of the Exception
     */
    public int requestPrivateChatCreate(int sock,
        boolean disposeWhenReplyReceived) throws IOException {

        HLProtocol.DataComponent[] dataComponents =
            new HLProtocol.DataComponent[]{
            hlp.new DataComponent(HLProtocol.HTLC_DATA_SOCKET, ToArrayConverters.intToByteArray(sock)),
            };

        HLProtocol.Packet packet =
            hlp.new Packet(HLProtocol.HTLC_HDR_PRIVCHAT_CREATE,
            nextTrans(),
            dataComponents);

        createTask(packet, disposeWhenReplyReceived);
        packet.write(output);

        return packet.header.trans;
    }

    /**
     *  Invite another user to join a private chat.
     *
     * @param  pcref            the reference of the private chat
     * @param  sock             the user to invite to join
     * @return                  transaction id
     * @exception  IOException  Description of the Exception
     */
    public int requestPrivateChatInvite(int pcref,
        int sock) throws IOException {

        return requestPrivateChatInvite(pcref, sock, true);
    }

    /**
     *  Description of the Method
     *
     * @param  pcref                     Description of the Parameter
     * @param  sock                      Description of the Parameter
     * @param  disposeWhenReplyReceived  Description of the Parameter
     * @return                           Description of the Return Value
     * @exception  IOException           Description of the Exception
     */
    public int requestPrivateChatInvite(int pcref,
        int sock,
        boolean disposeWhenReplyReceived) throws IOException {

        HLProtocol.DataComponent[] dataComponents =
            new HLProtocol.DataComponent[]{
            hlp.new DataComponent(HLProtocol.HTLC_DATA_SOCKET, 
                ToArrayConverters.intToByteArray(sock)),
            hlp.new DataComponent(HLProtocol.HTLC_DATA_CHAT_REF, 
                ToArrayConverters.intToByteArray(pcref)),
            };

        HLProtocol.Packet packet =
            hlp.new Packet(HLProtocol.HTLC_HDR_PRIVCHAT_INVITE,
            nextTrans(),
            dataComponents);

        createTask(packet, disposeWhenReplyReceived);
        packet.write(output);

        return packet.header.trans;
    }

    /**
     *  Send a request to join a private chat.
     *
     * @param  pcref            the reference of the private chat
     * @return                  transaction id
     * @exception  IOException  Description of the Exception
     */
    public int requestPrivateChatJoin(int pcref) throws IOException {
        return requestPrivateChatJoin(pcref, true);
    }

    /**
     *  Description of the Method
     *
     * @param  pcref                     Description of the Parameter
     * @param  disposeWhenReplyReceived  Description of the Parameter
     * @return                           Description of the Return Value
     * @exception  IOException           Description of the Exception
     */
    public int requestPrivateChatJoin(int pcref,
        boolean disposeWhenReplyReceived) throws IOException {

        HLProtocol.DataComponent[] dataComponents =
            new HLProtocol.DataComponent[]{
            hlp.new DataComponent(HLProtocol.HTLC_DATA_CHAT_REF, ToArrayConverters.intToByteArray(pcref)),
            };

        HLProtocol.Packet packet =
            hlp.new Packet(HLProtocol.HTLC_HDR_PRIVCHAT_JOIN,
            nextTrans(),
            dataComponents);

        createTask(packet, disposeWhenReplyReceived);
        packet.write(output);

        return packet.header.trans;
    }


    /**
     *  Decline an invitation to a private chat.
     *
     * @param  pcref            the reference of the private chat
     * @exception  IOException  Description of the Exception
     */
    public void sendPrivateChatDecline(int pcref) throws IOException {

        HLProtocol.DataComponent[] dataComponents =
            new HLProtocol.DataComponent[]{
            hlp.new DataComponent(HLProtocol.HTLC_DATA_CHAT_REF, ToArrayConverters.intToByteArray(pcref)),
            };

        HLProtocol.Packet packet =
            hlp.new Packet(HLProtocol.HTLC_HDR_PRIVCHAT_DECLINE,
            nextTrans(),
            dataComponents);

        packet.write(output);

    }


    /**
     *  Leaves a private chat.
     *
     * @param  pcref            the reference of the private chat
     * @exception  IOException  Description of the Exception
     */
    public void sendPrivateChatLeave(int pcref) throws IOException {

        HLProtocol.DataComponent[] dataComponents =
            new HLProtocol.DataComponent[]{
            hlp.new DataComponent(HLProtocol.HTLC_DATA_CHAT_REF, ToArrayConverters.intToByteArray(pcref)),
            };

        HLProtocol.Packet packet =
            hlp.new Packet(HLProtocol.HTLC_HDR_PRIVCHAT_LEAVE,
            nextTrans(),
            dataComponents);

        packet.write(output);

    }


    /**
     *  Sends text to a private chat.
     *
     * @param  pcref            the reference of the private chat
     * @param  s                the text to send
     * @exception  IOException  Description of the Exception
     */
    public void sendPrivateChat(int pcref,
        String s) throws IOException {

        if (s == null) {
            throw new IllegalArgumentException("s == null");
        }

        HLProtocol.DataComponent[] dataComponents =
            new HLProtocol.DataComponent[]{
                /* 2002.05.02 MackyBee modified */
                hlp.new DataComponent(HLProtocol.HTLC_DATA_CHAT, 
                // 2002.12.03 MackyBee modified. 
                s.getBytes(encoding)),
                //EncodingUtil.strictGetBytes(s)),
                //hlp.new DataComponent(HLProtocol.HTLC_DATA_CHAT, 
                //  s.getBytes()),
                hlp.new DataComponent(HLProtocol.HTLC_DATA_CHAT_REF, 
                    ToArrayConverters.intToByteArray(pcref)),
            };

        HLProtocol.Packet packet =
            hlp.new Packet(HLProtocol.HTLC_HDR_CHAT,
            nextTrans(),
            dataComponents);

        packet.write(output);

    }


    /**
     *  Sends an action to a private chat.
     *
     * @param  pcref            the reference of the private chat
     * @param  s                the action to send
     * @exception  IOException  Description of the Exception
     */
    public void sendPrivateChatAction(int pcref,
        String s) throws IOException {

        if (s == null) {
            throw new IllegalArgumentException("s == null");
        }

        HLProtocol.DataComponent[] dataComponents =
            new HLProtocol.DataComponent[]{
            hlp.new DataComponent(HLProtocol.HTLC_DATA_OPTION, ToArrayConverters.shortToByteArray(1)),
                /* 2002.05.02 MackyBee modified */
                hlp.new DataComponent(HLProtocol.HTLC_DATA_CHAT, 
                    // 2002.12.03 MackyBee modified. 
                    s.getBytes(encoding)),
                    //EncodingUtil.strictGetBytes(s)),
                //hlp.new DataComponent(HLProtocol.HTLC_DATA_CHAT, 
                //  s.getBytes()),
                hlp.new DataComponent(HLProtocol.HTLC_DATA_CHAT_REF, 
                    ToArrayConverters.intToByteArray(pcref)),
            };

        HLProtocol.Packet packet =
            hlp.new Packet(HLProtocol.HTLC_HDR_CHAT,
            nextTrans(),
            dataComponents);

        packet.write(output);

    }

    // 2002.11.11 MackyBee moved this into EncodingUtil class.
    // 2002.05.02 MackyBee added.
    /*
    public static byte[] strictGetBytes(String s) {
        try {
            return (new String(s.getBytes(
                pharerouge.client.Main.DEFAULT_ENCODING)).getBytes());
        } catch (UnsupportedEncodingException e) {
            DebuggerOutput.debug("HLClient: " + e.getMessage());
            // provisionally
            return (new byte[]{0});
        }
    }
    */

    /**
     *  Post a news article.
     *
     * @param  s                the news post
     * @return                  transaction id
     * @exception  IOException  Description of the Exception
     */
    public int requestNewsPost(String s) throws IOException {

        return requestNewsPost(s, true);
    }


    /**
     *  Description of the Method
     *
     * @param  s                         Description of the Parameter
     * @param  disposeWhenReplyReceived  Description of the Parameter
     * @return                           Description of the Return Value
     * @exception  IOException           Description of the Exception
     */
    public int requestNewsPost(String s,
        boolean disposeWhenReplyReceived) throws IOException {

        if (s == null) {
            throw new IllegalArgumentException("s == null");
        }

        HLProtocol.DataComponent[] dataComponents =
            new HLProtocol.DataComponent[]{
                /* 2002.05.02 MackyBee modified */
                hlp.new DataComponent(HLProtocol.HTLC_DATA_NEWS_POST, 
                    // 2002.12.03 MackyBee modified. 
                    s.getBytes(encoding))
                    //EncodingUtil.strictGetBytes(s))
                //hlp.new DataComponent(HLProtocol.HTLC_DATA_NEWS_POST, 
                // s.getBytes()),
            };

        HLProtocol.Packet packet =
            hlp.new Packet(HLProtocol.HTLC_HDR_NEWS_POST,
            nextTrans(),
            dataComponents);

        createTask(packet, disposeWhenReplyReceived);
        packet.write(output);

        return packet.header.trans;
    }

    /**
     *  Send a message to another user.
     *
     * @param  s                the message
     * @param  sock             Description of the Parameter
     * @return                  transaction id
     * @exception  IOException  Description of the Exception
     */
    public int requestMessageSend(int sock,
        String s) throws IOException {

        return requestMessageSend(sock, s, true);
    }

    /**
     *  Description of the Method
     *
     * @param  sock                      Description of the Parameter
     * @param  s                         Description of the Parameter
     * @param  disposeWhenReplyReceived  Description of the Parameter
     * @return                           Description of the Return Value
     * @exception  IOException           Description of the Exception
     */
    public int requestMessageSend(int sock,
        String s,
        boolean disposeWhenReplyReceived) throws IOException {

        if (s == null) {
            throw new IllegalArgumentException("s == null");
        }

        HLProtocol.DataComponent[] dataComponents =
            new HLProtocol.DataComponent[]{
            hlp.new DataComponent(HLProtocol.HTLC_DATA_SOCKET, 
                ToArrayConverters.intToByteArray(sock)),
            /* 2002.05.02 MackyBee modified */
            hlp.new DataComponent(HLProtocol.HTLC_DATA_MSG, 
                // 2002.12.03 MackyBee modified. 
                s.getBytes(encoding))
                //EncodingUtil.strictGetBytes(s)),
                //hlp.new DataComponent(HLProtocol.HTLC_DATA_MSG, 
                //s.getBytes()),
            };

        HLProtocol.Packet packet =
            hlp.new Packet(HLProtocol.HTLC_HDR_MSG,
            nextTrans(),
            dataComponents);

        createTask(packet, disposeWhenReplyReceived);
        packet.write(output);

        return packet.header.trans;
    }

    /**
     *  Kick a user.
     *
     * @param  sock             the user to kick
     * @return                  transaction id
     * @exception  IOException  Description of the Exception
     */
    public int requestUserKick(int sock) throws IOException {
        return requestUserKick(sock, true);
    }

    /**
     *  Description of the Method
     *
     * @param  sock                      Description of the Parameter
     * @param  disposeWhenReplyReceived  Description of the Parameter
     * @return                           Description of the Return Value
     * @exception  IOException           Description of the Exception
     */
    public int requestUserKick(int sock,
        boolean disposeWhenReplyReceived) throws IOException {

        HLProtocol.DataComponent[] dataComponents =
            new HLProtocol.DataComponent[]{
            hlp.new DataComponent(HLProtocol.HTLC_DATA_SOCKET, ToArrayConverters.intToByteArray(sock)),
            };

        HLProtocol.Packet packet =
            hlp.new Packet(HLProtocol.HTLC_HDR_USER_KICK,
            nextTrans(),
            dataComponents);

        createTask(packet, disposeWhenReplyReceived);
        packet.write(output);

        return packet.header.trans;
    }

    /**
     *  Kicks and bans a user.
     *
     * @param  sock             the user to ban
     * @return                  transaction id
     * @exception  IOException  Description of the Exception
     */
    public int requestUserKickBan(int sock) throws IOException {

        return requestUserKickBan(sock, true);
    }

    /**
     *  Description of the Method
     *
     * @param  sock                      Description of the Parameter
     * @param  disposeWhenReplyReceived  Description of the Parameter
     * @return                           Description of the Return Value
     * @exception  IOException           Description of the Exception
     */
    public int requestUserKickBan(int sock,
        boolean disposeWhenReplyReceived) throws IOException {

        HLProtocol.DataComponent[] dataComponents =
            new HLProtocol.DataComponent[]{
            hlp.new DataComponent(HLProtocol.HTLC_DATA_BAN, 
                ToArrayConverters.shortToByteArray(1)),
            hlp.new DataComponent(HLProtocol.HTLC_DATA_SOCKET, 
                ToArrayConverters.intToByteArray(sock)),
            };

        HLProtocol.Packet packet =
            hlp.new Packet(HLProtocol.HTLC_HDR_USER_KICK,
            nextTrans(),
            dataComponents);

        createTask(packet, disposeWhenReplyReceived);
        packet.write(output);

        return packet.header.trans;
    }

    /* Utility methods */
    /**
     *  Waits for the specified task to complete and returns it's result. For
     *  this to work <i>reliably</i> , the task that is being waited for <i>must
     *  </i> have been created with the <tt>disposeWhenReplyReceived</tt> flag
     *  set to <tt>false</tt> . If the task number does not exist, returns null.
     *
     * @param  id                         the transaction id to wait for.
     * @return                            a data object
     * @exception  HLException            Description of the Exception
     * @exception  IOException            Description of the Exception
     * @exception  InterruptedException   Description of the Exception
     * @throws  IllegalArgumentException  when the task specified by id has
     *      automatic dispose turned on.
     */
    public Object waitFor(int id) throws HLException, IOException, 
        InterruptedException {

        Object o = null;

        try {

            HLTask task = getTask(id);

            if (task.disposeWhenReceived) {
                throw new IllegalArgumentException(
                    "aborted attempted wait for task that has "
                    + "disposeWhenReceived flag set to true; unwise.");
            }

            o = task.getData();

            DebuggerOutput.debug("HLClient.waitFor[" + id + "]: got data.");

        } finally {

            //            if(lockTask(id)) {

            //                DebuggerOutput.debug("HLClient.waitFor[" + id + 
            //"]: disposing task[" + id + "].");
            
            disposeTask(id);

            //            } else {

            //                DebuggerOutput.debug("HLClient.waitFor[" + id + 
            // "]: task locked, won't dispose.");

            //            }
        }

        return o;
    }

    /**
     *  Returns a string representation of this object.
     *
     * @return    a string representation.
     */
    // 2002.10.21 MackyBee EXPERIMENTALLY deactivated synchronized.
    public /*synchronized*/ String toString() {
        String s = "HLClient";
        if (host != null) {
            s += host.toString();
            s += connected ? " [connected]" : " [disconnected]";

            if (connected) {
                s += " tasks: " + tasks;
            }
        } else {
            s += " no host";
        }
        return s;
    }

    /* Modifier methods */
    /**
     *  Sets the server version.
     *
     * @param  v  the version.
     */
    void setServerVersion(int v) {
        serverVersion = v;
    }

    /**
     *  This is a dummy method to let > 1.5 servers know that the agreement has
     *  been accepted or something...
     *
     * @exception  IOException  Description of the Exception
     */
    void sendAgree() throws IOException {
        //{{{ deactivated process.
        /*
			Experimentally disabled since 2002.02.25 by MackyBee
			Uses sendLoginUserInfo method instead.of this.
		*/
        /*
        HLProtocol.DataComponent[] dataComponents =
            new HLProtocol.DataComponent[] {
                hlp.new DataComponent(HLProtocol.HTLC_DATA_NICK, 
                //nick.getBytes()),
                hlp.new DataComponent(HLProtocol.HTLC_DATA_ICON, 
                //ToArrayConverters.charToByteArray((char) icon)),
            };
        */
        /* fixme: Must do a createTask() here. */
        /*
        HLProtocol.Packet packet =
            hlp.new Packet(HLProtocol.HTLC_HDR_AGREE,
                           nextTrans(),
                           null);
        packet.write(output);
		*/
        //}}}
    }

    /**
     *  Description of the Method
     *
     * @param  hdr              Description of the Parameter
     * @exception  IOException  Description of the Exception
     */
    // 2003.02.11 MackyBee modified. 
    /* Added exprerimentally 2002.02.25 by MackyBee */
    private void sendLoginUserInfo(int hdr) throws IOException {

        /* Added experimentally 2002.03.06 MackyBee */
        /* Modified the following note 2002.03.08 MackyBee */
        /* 	This data specifies client status which "Private Chat" and
			"Private Message" availability.
			As far as I know,
			{0, 1} means that user isn't allowed Private Message. */
        /* Now, this is not a "unknown" */
        byte[] unknownData = {0, 0};

        // 2002.09.30 MackyBee added.
        HLProtocol.DataComponent[] dataComponents =
            new HLProtocol.DataComponent[(getServerVersion() < 150) ? 2 : 3];

        // 2002.09.30 MackyBee modified.
        dataComponents[0] = 
        //HLProtocol.DataComponent[] dataComponents =
        //    new HLProtocol.DataComponent[] {
            hlp.new DataComponent(HLProtocol.HTLC_DATA_NICK,
            // 2002.12.03 MackyBee modified. 
            // 2002.05.02 MackyBee modified
            nick.getBytes(encoding));
            //EncodingUtil.strictGetBytes(nick));
            //,
            //nick.getBytes()), 
        dataComponents[1] = 
            hlp.new DataComponent(HLProtocol.HTLC_DATA_ICON,
            ToArrayConverters.shortToByteArray(icon));
            //, 
        // 2002.10.22 MackyBee tuned this if condition.
        if (dataComponents.length > 2) {
        //if (!(getServerVersion() < 150)) {
            dataComponents[2] = 
            // 2002.10.22 MackyBee replaced raw value to constant.
            // Added experimentally 2002.03.06 MackyBee
            // As HLC 1.8.5
            hlp.new DataComponent(HLProtocol.HTLC_DATA_COMM_RESTRICTION, 
                unknownData);
        }
        //};

        HLProtocol.Packet packet =
            hlp.new Packet(hdr,
            nextTrans(),
            dataComponents);

        /* Experimentally disabled 2002.03.03 MackyBee */
        /* This process doesn't need making task, I think so. */
        //HLTask task = createTask(packet, false);

        DebuggerOutput.debug("HLClient: writing user informations");

        packet.write(output);
    }

    /**
     *  Invokes handleDisconnect() on registered HLClientListeners.
     *
     * @param  msg  Description of the Parameter
     */
    void fireDisconnect(String msg) {
        DebuggerOutput.debug("HLClient.fireDisconnect: accepted (reason = " 
            + msg + ")");

        for (Enumeration en = hle.elements(); en.hasMoreElements(); ) {
            HLClientListener l = (HLClientListener) en.nextElement();
            DebuggerOutput.debug(
                "HLClient.fireDisconnect: invoking handleDisconnect on " + l);
            l.handleDisconnect(msg);
        }
    }

     /* Following methods are about task handling. Might as well make
       this into a separate class. */

    /**
     *  Description of the Method
     *
     * @param  msg  Description of the Parameter
     */
    void terminateAllTasks(String msg) {
        // 2003.02.22 MackyBee modified. 
        // It had been occurred a ConcurrentModificationException here. 
        // 2002.10.25 MackyBee modified for spec change.
        Object[] keys = tasks.keySet().toArray();
        //Iterator k = tasks.keySet().iterator();
        //Enumeration k = tasks.keys();

        // 2003.02.22 MackyBee added. 
        int trans = 0;
        
        // 2002.10.25 MackyBee modified for spec change.
        for (int k = 0; k < keys.length; k++) {
        //while (k.hasNext()) {
        //while (k.hasMoreElements()) {

            trans = ((Integer) keys[k]).intValue();
            //int trans = ((Integer) k.next()).intValue();
            //int trans = ((Integer) k.nextElement()).intValue();

            if (lockTask(trans)) {
                /* If we can't lock the task, then the dispatcher
                   got to this task first. That's OK. Just let it
                   process the task reply. */
                try {
                    DebuggerOutput.debug("terminateAllTasks: terminating task " 
                        + getTask(trans).toString());

                    HLTask task = getTask(trans);
                    task.setError(new HLException(msg));

                    /* Have to dispose this task here because the
                       dispatcher won't be able to. */
                    if (task.disposeWhenReceived) {
                        disposeTask(trans);
                    }
                } catch (HLTaskNotFoundException e) {
                    DebuggerOutput.stackTrace(e);
                } catch (HLException e) {
                    // 2002.09.29 MackyBee added this block.
                    DebuggerOutput.stackTrace(e);
                }
            }
        }
    }

    /**
     *  Description of the Method
     *
     * @param  trans  Description of the Parameter
     * @return        Description of the Return Value
     */
    synchronized boolean lockTask(int trans) {
        boolean lockSucceeded = false;
        HLTask task;

        /* It's possible that the task was disposed before we could
           get to it. That's not possible here, because disposeTask()
           is synchronized. */
        if (!tasks.containsKey(new Integer(trans))) {

            DebuggerOutput.debug("lockTask[" + trans + "]: task does not exist");
            return false;
        }

        try {

            task = getTask(trans);

            if (!task.isLocked()) {

                DebuggerOutput.debug("lockTask[" + trans + "]: task not locked, locking task ...");
                task.lock();
                DebuggerOutput.debug("lockTask[" + trans + "]: task locked.");
                lockSucceeded = true;

            } else {

                DebuggerOutput.debug("lockTask[" + trans + "]: task already locked, not locking");

            }

        } catch (HLTaskNotFoundException e) {
            DebuggerOutput.stackTrace(e);
        }

        return lockSucceeded;
    }

    /**
     *  Description of the Method
     *
     * @return    Description of the Return Value
     */
    // 2002.10.14 MackyBee deactivated modifier "synchronized" because this 
    // int increment process is atomic.
    /* synchronized */ int nextTrans() {
        return transaction++;
    }

    /**
     *  Description of the Method
     *
     * @param  packet                    Description of the Parameter
     * @param  disposeWhenReplyReceived  Description of the Parameter
     * @return                           Description of the Return Value
     * @exception  IOException           Description of the Exception
     */
    // 2002.10.21 MackyBee EXPERIMENTALLY deactivated synchronized.
    /*synchronized*/ HLTask createTask(HLProtocol.Packet packet,
        boolean disposeWhenReplyReceived) throws IOException {

        if (!connected) {
            throw new IOException("not connected");
        }

        HLTask task = new HLTask(packet.header.id, disposeWhenReplyReceived);
        tasks.put(new Integer(packet.header.trans), task);
        DebuggerOutput.debug("createTask: created task[" 
            + packet.header.trans + "]: " + task);

        return task;
    }

    /**
     *  Description of the Method
     *
     * @param  trans                        Description of the Parameter
     * @exception  HLTaskNotFoundException  Description of the Exception
     */
    // 2002.10.21 MackyBee EXPERIMENTALLY deactivated synchronized.
    /*synchronized*/ void disposeTask(int trans) 
        throws HLTaskNotFoundException {

        DebuggerOutput.debug("disposeTask: disposing task [" + trans + "] ... ");
        DebuggerOutput.debug("disposeTask: task list = " + tasks);

        Transfer xf = getTransfer(trans);

        if (xf != null) {
            disposeTransfer(trans);
        }

        if (tasks.containsKey(new Integer(trans))) {
            DebuggerOutput.debug("disposeTask: disposed task [" + trans + "].");
            tasks.remove(new Integer(trans));
        } else {
            DebuggerOutput.debug("disposeTask: no such task [" + trans + "]");
            throw new HLTaskNotFoundException("no such task: " + trans);
        }
    }

    /**
     *  Gets the task attribute of the HLClient object
     *
     * @param  trans                        Description of the Parameter
     * @return                              The task value
     * @exception  HLTaskNotFoundException  Description of the Exception
     */
    // 2002.10.21 MackyBee EXPERIMENTALLY deactivated synchronized.
    /*synchronized*/ HLTask getTask(int trans) throws HLTaskNotFoundException 
        {

        if (tasks.containsKey(new Integer(trans))) {
            return (HLTask) tasks.get(new Integer(trans));
        } else {
            throw new HLTaskNotFoundException("no such task: " + trans);
        }
    }

    /**
     *  Description of the Method
     *
     * @param  packet  Description of the Parameter
     * @param  tx      Description of the Parameter
     */
    // 2002.10.21 MackyBee EXPERIMENTALLY deactivated synchronized.
    /*synchronized*/ void createTransfer(HLProtocol.Packet packet, Transfer tx)
        {
        createTransfer(packet.header.trans, tx);
    }

    /**
     *  Description of the Method
     *
     * @param  trans  Description of the Parameter
     * @param  tx     Description of the Parameter
     */
    // 2002.10.21 MackyBee EXPERIMENTALLY deactivated synchronized.
    /*synchronized*/ void createTransfer(int trans, Transfer tx) {
        DebuggerOutput.debug("HLClient.createTransfer: transfer = " + trans);
        transfers.put(new Integer(trans), tx);
    }

    /**
     *  Description of the Method
     *
     * @param  trans  Description of the Parameter
     */
    // 2002.10.21 MackyBee EXPERIMENTALLY deactivated synchronized.
    /*synchronized*/ void disposeTransfer(int trans) {
        Transfer xf = (Transfer) transfers.get(new Integer(trans));

        if (xf != null) {
//            if (xf.local != null) {
//                try {
//                    xf.local.close();
//                } catch (IOException e) {
//                    DebuggerOutput.stackTrace(e);
//                }
//            }

            /*
            if(xf.inputPipe != null) {
                try {
                    //                    xf.inputPipe.close();
                } catch(IOException e) {
                    DebuggerOutput.stackTrace(e);
                }
            }
            */
            transfers.remove(new Integer(trans));
        }
    }

    /**
     *  Gets the transfer attribute of the HLClient object
     *
     * @param  trans  Description of the Parameter
     * @return        The transfer value
     */
    // 2002.10.21 MackyBee EXPERIMENTALLY deactivated synchronized.
    /*synchronized*/ Transfer getTransfer(int trans) {
        return (Transfer) transfers.get(new Integer(trans));
    }

    /**
     *  Description of the Method
     *
     * @exception  IOException  Description of the Exception
     */
    private synchronized void socketClose() throws IOException {

        if (socket != null) {
            DebuggerOutput.debug("HLClient.close: closing socket");
            socket.close();
            socket = null;
        }

        /* InterruptableStream overrides close to interrupt any
           pending reads. */
        if (input != null) {
            DebuggerOutput.debug("HLClient.close: closing input");
            input.close();
            input = null;
        }
    }

    /**
     *  Description of the Method
     *
     * @param  path                      Description of the Parameter
     * @param  local                     Description of the Parameter
     * @param  inputPipe                 Description of the Parameter
     * @param  meter                     Description of the Parameter
     * @param  forkOrFile                Description of the Parameter
     * @param  disposeWhenReplyReceived  Description of the Parameter
     * @return                           Description of the Return Value
     * @exception  IOException           Description of the Exception
     */
//    private int requestForkOrFileDownload(String path,
//        MacFile local,
//        PipedInputStream inputPipe,
//        Meter meter,
//        int forkOrFile,
//        boolean disposeWhenReplyReceived) throws IOException {
//
//        if (meter == null) {
//            throw new IllegalArgumentException("meter object cannot be null");
//        }
//
//        if (path == null) {
//            throw new IllegalArgumentException("null argument");
//        }
//
//        HLProtocol.DataComponent[] pathComponents =
//            // 2002.12.22 MackyBee modified. 
//            hlp.createPathComponents(
//            //ComponentFactory.createPathComponents(
//            path,
//            HLProtocol.HTLC_DATA_DIR,
//            true);
//
//        HLProtocol.DataComponent[] dataComponents =
//            new HLProtocol.DataComponent[]{
//            hlp.new DataComponent(HLProtocol.HTLC_DATA_RESUME,
//            ToArrayConverters.shortToByteArray(2)),
//            pathComponents[1],
//            pathComponents[0],
//            };
//
//        if (forkOrFile != HLProtocol.DL_DATA_FORK) {
//            dataComponents[0] = hlp.new ResumeTransferComponent(
//                local.getDataFork().size(), local.getResourceFork().size());
//        }
//
//        HLProtocol.Packet packet =
//            hlp.new Packet(HLProtocol.HTLC_HDR_FILE_GET,
//            nextTrans(),
//            dataComponents);
//
//        Transfer tx = new Transfer();
//        tx.name = path;
//        tx.local = local;
//        tx.inputPipe = inputPipe;
//        tx.remote = path;
//
//        if (forkOrFile != HLProtocol.DL_DATA_FORK) {
//            tx.local_data_size = (int) local.getDataSize();
//            tx.local_rsrc_size = (int) local.getResourceSize();
//        }
//
//        tx.remote_data_size = tx.remote_rsrc_size = 0;
//        tx.meter = meter;
//        tx.forkOrFile = forkOrFile;
//
//        createTask(packet, disposeWhenReplyReceived);
//        createTransfer(packet, tx);
//        packet.write(output);
//
//        return packet.header.trans;
//    }

    /**
     *  Moves or links a file.
     *
     * @param  kind                      HLProtocol.HTLC_HDR_FILE_MAKE_ALIAS or
     *      HLProtocol.HTLC_HDR_FILE_MOVE.
     * @param  source                    the source file.
     * @param  target                    the target file.
     * @param  disposeWhenReplyReceived  Description of the Parameter
     * @return                           Description of the Return Value
     * @exception  IOException           Description of the Exception
     */
    private int requestFileMoveOrMakeAlias(int kind,
        String source,
        String target,
        boolean disposeWhenReplyReceived) throws IOException {

        if (source == null || target == null) {
            throw new IllegalArgumentException("null argument");
        }

        HLProtocol.DataComponent[] sourcePathComponents =
            // 2002.12.22 MackyBee modified. 
            hlp.createPathComponents(
            //ComponentFactory.createPathComponents(
            source,
            HLProtocol.HTLC_DATA_DIR,
            true);

        HLProtocol.DataComponent[] targetPathComponents =
            // 2002.12.22 MackyBee modified. 
            hlp.createPathComponents(
            //ComponentFactory.createPathComponents(
            target,
            HLProtocol.HTLC_DATA_DIR_RENAME,
            false);

        HLProtocol.DataComponent[] dataComponents =
            new HLProtocol.DataComponent[]{
            sourcePathComponents[1],
            sourcePathComponents[0],
            targetPathComponents[0],
            };

        HLProtocol.Packet packet =
            hlp.new Packet(kind,
            nextTrans(),
            dataComponents);

        createTask(packet, disposeWhenReplyReceived);
        packet.write(output);

        return packet.header.trans;
    }

    /**
     *  Internal method for modifying / creating a user account.
     *
     * @param  login                     Description of the Parameter
     * @param  password                  Description of the Parameter
     * @param  nick                      Description of the Parameter
     * @param  privileges                Description of the Parameter
     * @param  packetType                Description of the Parameter
     * @param  disposeWhenReplyReceived  Description of the Parameter
     * @return                           Description of the Return Value
     * @exception  IOException           Description of the Exception
     */
    private int requestAccountCreateOrModify(String login,
        String password,
        String nick,
        long privileges,
        int packetType,
        boolean disposeWhenReplyReceived) throws IOException {

        if (login == null || nick == null) {
            throw new IllegalArgumentException("null argument");
        }

        // 2002.10.28 MackyBee notes: 
        // Official HLC/HLS treats 1 byte 0 (zero) password data as 
        // "no password change". So, modified the treat on UserEditInterface.
        // 2002.10.27 MackyBee IMPORTANT Notes: 
        // If password == null, it should not send any password data packet.
        // This process is detected as "no password change" by server. NO!

        // 2002.10.27 MackyBee modified.
        byte[] p = null;
        if (password != null) {
            // 2002.10.28 MackyBee added this if statement block.
            if (password.equals("")) {
                // This process refers to official HLC behaviour.
                p = new byte[1];
                p[0] = 0;
            } else {
                // 2002.12.03 MackyBee modified. 
                p = HLProtocol.invert(password.getBytes(encoding));
                //p = HLProtocol.invert(EncodingUtil.strictGetBytes(password));
            }
        }
        /*
        byte[] p = (password == null ?
            new byte[1] : 
            // 2002.05.02 MackyBee modified.
            HLProtocol.invert(strictGetBytes(password)));
            //HLProtocol.invert(password.getBytes()));
        */

        // 2002.10.27 MackyBee replaced the following process.
        int dcSize = 4;
        if (password == null) {
            dcSize--;
        }

        HLProtocol.DataComponent[] dataComponents = 
            new HLProtocol.DataComponent[dcSize];

        int dcIndex = 0;

        dataComponents[dcIndex] = hlp.new DataComponent(
            HLProtocol.HTLC_DATA_LOGIN, 
            // 2002.12.03 MackyBee modified. 
            HLProtocol.invert(login.getBytes(encoding)));
            //HLProtocol.invert(login.getBytes()));
        dcIndex++;
        if (password != null) {
            dataComponents[dcIndex] = hlp.new DataComponent(
                HLProtocol.HTLC_DATA_PASSWORD, p);
            dcIndex++;
        }
        dataComponents[dcIndex] = hlp.new DataComponent(
            HLProtocol.HTLC_DATA_NICK, 
            // 2002.12.03 MackyBee modified. 
            nick.getBytes(encoding));
            //EncodingUtil.strictGetBytes(nick));
        dcIndex++;
        dataComponents[dcIndex] = hlp.new DataComponent(
            HLProtocol.HTLS_DATA_PRIVILEGES, 
            ToArrayConverters.swapByteArray(
            ToArrayConverters.longToByteArray(privileges)));
        //{{{ deactivated lines
        /*
        HLProtocol.DataComponent[] dataComponents =
            new HLProtocol.DataComponent[]{
            hlp.new DataComponent(HLProtocol.HTLC_DATA_LOGIN, 
                HLProtocol.invert(login.getBytes())),
            hlp.new DataComponent(HLProtocol.HTLC_DATA_PASSWORD, p), 
            // 2002.05.02 MackyBee modified.
            hlp.new DataComponent(HLProtocol.HTLC_DATA_NICK, 
                strictGetBytes(nick)),
            //hlp.new DataComponent(HLProtocol.HTLC_DATA_NICK, nick.getBytes()),
            hlp.new DataComponent(HLProtocol.HTLS_DATA_PRIVILEGES, 
                ToArrayConverters.swapByteArray(
                ToArrayConverters.longToByteArray(privileges))),
            };
        */
        //}}}

        HLProtocol.Packet packet =
            hlp.new Packet(packetType,
            nextTrans(),
            dataComponents);

        createTask(packet, disposeWhenReplyReceived);
        packet.write(output);

        return packet.header.trans;
    }

    /**  Description of the Method */
    private synchronized void internalDisconnect() {

        if (!connected) {
            DebuggerOutput.debug("HLClient.internalDisconnect: "
                + "already disconnected.");
            return;
        }

        /* Wake up any threads still waiting for an agreement or a
           task reply. */
        hlr.notifyAgreementAborted();

        DebuggerOutput.debug("HLClient.internalDisconnect: terminating tasks");
        terminateAllTasks("connection closed");
        DebuggerOutput.debug("HLClient.internalDisconnect: tasks terminated");

        connected = false;

        // 2002.09.29 MackyBee modified.
        transaction = 1;
        //transaction = 2;
    }
}

//{{{ class FileTransferThread
/**  Description of the Class */
class FileTransferThread extends Thread {

    HLProtocol.Packet packet;
    HLClient hlc;

    /**
     *  Constructor for the FileTransferThread object
     *
     * @param  hlc     Description of the Parameter
     * @param  packet  Description of the Parameter
     */
    FileTransferThread(HLClient hlc, HLProtocol.Packet packet) {

        this.hlc = hlc;
        this.packet = packet;

        setName("FileTransferThread[task[" + packet.header.trans + "]]");
        setPriority(Thread.MIN_PRIORITY);

    }

    /**  Main processing method for the FileTransferThread object */
    public void run() {
        hlc.hlr.dispatch(packet);
    }

}
//}}}
